
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from ..Mint import Mint

from . import tools as T

scaling_options = [{'value': i, 'label': i} for i in ['Standard']]

_layout = html.Div([
    html.H3('Decomposition'),
    html.Button('Run PCA',id='pca-update'),
    dcc.Dropdown(id='dec-scaling', options=scaling_options, 
        value=['Standard'], multi=True, placeholder='Scaling used before PCA'),

    html.Div(id='pca-figures', style={'margin': 'auto', 'text-align': 'center'})
    
])

_label = 'PCA'

def layout():
    return _layout


def callbacks(app, fsc, cache):
    
    @app.callback(
        Output('pca-figures', 'children'),
        Input('pca-update', 'n_clicks'),
        State('ana-groupby', 'value'),
        State('ana-peak-labels-include', 'value'),
        State('ana-peak-labels-exclude', 'value'),
        State('ana-file-types', 'value'),
        State('wdir', 'children')
    )
    def create_pca( n_clicks, groupby, include_labels, exclude_labels, 
            file_types, wdir ):
        if n_clicks is None:
            raise PreventUpdate

        df = T.get_complete_results( wdir, include_labels=include_labels, 
            exclude_labels=exclude_labels, file_types=file_types )

        if file_types is not None and len(file_types) > 0:
            df = df[df['Type'].isin(file_types)]
        if groupby is not None and len(groupby) > 0:
            color_groups = df[['ms_file', groupby]].drop_duplicates().set_index('ms_file')
            #ndx_cgrp = color_groups.index.to_list()
        else:
            color_groups = None
            groupby = None

        figures = []
        mint = Mint()
        mint.results = df
        mint.pca()

        ndx = mint.decomposition_results['df_projected'].index.to_list()

        mint.pca_plot_cummulative_variance()
        src = T.fig_to_src()
        figures.append( html.Img(src=src) )

        if color_groups is not None:
            color_groups = color_groups.loc[ndx].values

        mint.plot_pair_plot(group_name=groupby, 
                color_groups=color_groups)

        src = T.fig_to_src()
        figures.append( html.Img(src=src) )

        return figures


import pandas as pd

class Resampler():
    def __init__(self, frequeny='50ms', unit='minute'):
        self._unit = unit
        self._frequency = frequeny
    
    def resample_unit_minutes(self, chrom):
        chrom = chrom.sort_index()
        chrom.index = pd.to_timedelta(chrom.index, unit='minute')
        chrom = chrom.resample(self._frequency).nearest()
        chrom.index = (chrom.index.seconds + chrom.index.microseconds / 1000000) / 60
        chrom = chrom.rolling(20, center=True).mean()
        chrom = chrom.rolling(5, center=True).mean()
        return chrom
    
    def resample(self, chrom):
        if self._unit == 'minute':
            return self.resample_unit_minutes(chrom)
        else:
            raise NotImplementedError


import os
import shutil
import uuid
import logging

import wget
import urllib3, ftplib
from urllib.parse import urlparse
from bs4 import BeautifulSoup

from glob import glob

import pandas as pd

import dash_bootstrap_components as dbc
import dash_html_components as html
import dash_core_components as dcc

from dash.exceptions import PreventUpdate

from dash.dependencies import Input, Output, State

from ms_mint.io import convert_ms_file_to_feather

from dash_tabulator import DashTabulator
from tqdm import tqdm

import dash_uploader as du



from . import tools as T


options = {
           "selectable": True,
           "headerFilterLiveFilterDelay":3000,
           "layout": "fitDataFill",
           "height": "900px",
           }

clearFilterButtonType = {"css": "btn btn-outline-dark", "text":"Clear Filters"}


# If color column is not present, for some strange reason, the header filter disappears.
columns = [
        { "formatter": "rowSelection", "titleFormatter":"rowSelection",  
          "titleFormatterParams": {
             "rowRange": "active" # only toggle the values of the active filtered rows
          },
          "hozAlign":"center", "headerSort": False, "width":"1px", 'frozen': True},
        { "title": "MS-file", "field": "MS-file", "headerFilter":True, 
          'headerSort': True, "editor": "input", "width": "100%",
          'sorter': 'string', 'frozen': True},
        { 'title': '', 'field': '', "headerFilter":False,  "formatter":"color", 
          'width': '3px', "headerSort": False},
    ]


ms_table = html.Div(id='ms-table-container', 
    style={'minHeight':  100, 'marginTop': '10%'},
    children=[
        DashTabulator(id='ms-table',
            columns=columns, 
            options=options,
            clearFilterButtonType=clearFilterButtonType
        )
])

_label = 'MS-Files'

_layout = html.Div([
    html.H3('Upload MS-files'),
    dcc.Upload(
            id='ms-upload',
            children=html.Div([
                html.A('''Click here or drag and drop to upload mzML/mzXML files here. 
                You can upload up to 10 files at a time.''', style={'margin': 'auto', 'padding': 'auto'})
            ]),
            style={
                'width': '100%',
                'height': '120px',
                'lineHeight': '120px',
                'borderWidth': '1px',
                'borderStyle': 'dashed',
                'borderRadius': '5px',
                'textAlign': 'center',
                'marginBottom': '15px',
            },
            # Allow multiple files to be uploaded
            multiple=True),
    html.Div( du.Upload(id='ms-upload-zip', filetypes=['tar', 'zip'], 
                        upload_id=uuid.uuid1(),
                        text='Click here to drag and drop ZIP/TAR compressed archives'),
              style={
                    'textAlign': 'center',
                    'width': '100%',
                    'padding': '0px',
                    'display': 'inline-block',
                }),
    html.Button('Import from URL', id='ms-import-from-url'),
    dcc.Input(id='url', placeholder='Drop URL here', style={'width': '100%'}),
    dcc.Markdown('---', style={'marginTop': '10px'}),
    dcc.Markdown('##### Actions'),
    html.Button('Convert to Feather', id='ms-convert'),
    html.Button('Delete selected files', id='ms-delete', style={'float': 'right'}),
    dcc.Loading( ms_table ),
    html.Div(id='ms-upload-zip-filename', style={'visibility': 'hidden'}),    
])


_outputs = html.Div(id='ms-outputs', 
    children=[
        html.Div(id={'index': 'ms-upload-output', 'type': 'output'}),
        html.Div(id={'index': 'ms-convert-output', 'type': 'output'}),
        html.Div(id={'index': 'ms-delete-output', 'type': 'output'}),
        html.Div(id={'index': 'ms-save-output', 'type': 'output'}),
        html.Div(id={'index': 'ms-import-from-url-output', 'type': 'output'}),
        html.Div(id={'index': 'ms-upload-zip-output', 'type': 'output'}),
    ]
)


def layout():
    return _layout


def callbacks(app, fsc, cache):

    @app.callback(
    Output({'index': 'ms-upload-output', 'type': 'output'}, 'children'),
    Input('ms-upload', 'contents'),
    Input({'index':'ms-convert-output', 'type': 'output'}, 'children'),
    State('ms-upload', 'filename'),
    State('ms-upload', 'last_modified'),
    State('wdir', 'children'))
    def ms_upload(list_of_contents, converted, list_of_names, list_of_dates, wdir):
        target_dir = os.path.join(wdir, 'ms_files')
        if list_of_contents is not None:
            n_total = len(list_of_contents)
            n_uploaded = 0
            for i, (c, n, d) in enumerate( zip(list_of_contents, list_of_names, list_of_dates) ):
                fsc.set('progress', int( 100*(i+1)/n_total ))
                if n.lower().endswith('mzxml') or n.lower().endswith('mzml') or n.lower().endswith('zip'):
                    try:
                        T.parse_ms_files(c, n, d, target_dir)
                        n_uploaded += 1
                    except:
                        logging.warning(f'Could not parse file {n}')
                if n.lower().endswith('zip'):
                    logging.info(f'Zip file uploaded {target_dir}, {n}')
                    fn = os.path.join( target_dir, n)
                    shutil.unpack_archive(fn, target_dir)
                    os.remove(fn)
            return dbc.Alert(f'{n_uploaded} files uploaded.', color='success')


    @app.callback(
    Output('ms-table', 'data'),
    Input({'index': 'ms-upload-output', 'type': 'output'}, 'children'),
    Input('wdir', 'children'), 
    Input({'index':'ms-delete-output', 'type': 'output'}, 'children'),
    Input({'index':'ms-upload-zip-output', 'type': 'output'}, 'children')
    )
    def ms_table(value, wdir, files_deleted, zip_extracted): 
        ms_files = T.get_ms_fns( wdir )
        data = pd.DataFrame({'MS-file':  [os.path.basename(fn) for fn in ms_files],
                             'Size[MB]': [os.path.getsize(fn)  for fn in ms_files]})
        return data.to_dict('records')


    @app.callback(
    Output({'index': 'ms-convert-output', 'type': 'output'}, 'children'),
    Input('ms-convert', 'n_clicks'),
    State('ms-table', 'multiRowsClicked'),
    State('wdir', 'children')
    )
    def ms_convert(n_clicks, rows, wdir):
        target_dir = os.path.join(wdir, 'ms_files')
        if n_clicks is None:
            raise PreventUpdate
        fns = [row['MS-file'] for row in rows]
        fns = [fn for fn in fns if not fn.endswith('.feather')]
        fns = [os.path.join(target_dir, fn) for fn in fns]
        n_total = len(fns)
        for i, fn in enumerate( fns ):
            fsc.set('progress', int(100*(i+1)/n_total))
            new_fn = convert_ms_file_to_feather(fn)
            if os.path.isfile(new_fn): os.remove(fn)
        return dbc.Alert('Files converted to feather format.', color='info')


    @app.callback(
        Output({'index': 'ms-delete-output', 'type': 'output'}, 'children'),
        Input('ms-delete', 'n_clicks'),
        State('ms-table', 'multiRowsClicked'),
        State('wdir', 'children')
    )
    def ms_delete(n_clicks, rows, wdir):
        if n_clicks is None:
            raise PreventUpdate
        target_dir = os.path.join(wdir, 'ms_files')
        for row in rows:
            fn = row['MS-file']
            fn = os.path.join(target_dir, fn)
            os.remove(fn)
        return dbc.Alert(f'{len(rows)} files deleted', color='info')


    @du.callback(
    output=Output('ms-upload-zip-filename', 'children'),
    id='ms-upload-zip',
    )
    def get_zip_filename(filenames):
        return filenames[0]

    @app.callback(
        Output({'index': 'ms-upload-zip-output', 'type': 'output'}, 'children'),
        Input('ms-upload-zip-filename', 'children'),
        State('wdir', 'children')
    )
    def get_a_list(fn, wdir):
        if fn is None:
            raise PreventUpdate
        ms_dir = T.get_ms_dirname( wdir )
        upload_path = os.path.dirname( fn )
        shutil.unpack_archive(fn, extract_dir=upload_path)
        search_pattern = os.path.join( upload_path, '**', '*.*')
        fns = glob(search_pattern, recursive=True)
        n_total = len(fns)
        for i, fn in tqdm( enumerate(fns), total=n_total ):
            fsc.set('progress', int( 100 * (i+1) / n_total))
            if fn.lower().endswith('mzxml') or fn.lower().endswith('mzml'):
                try:
                    shutil.move(fn, ms_dir)
                except:
                    pass
        for remainings in glob(os.path.join(upload_path, '*')):
            if os.path.isfile(remainings): os.remove(remainings)
            elif os.path.isdir(remainings): shutil.rmtree(remainings)
        return dbc.Alert('Upload finished', color='success')


    @app.callback(
        Output({'index': 'ms-import-from-url-output', 'type': 'output'}, 'children'),
        Input('ms-import-from-url', 'n_clicks'),
        State('url', 'value'),
        State('wdir', 'children')
    )
    def import_from_url(n_clicks, url, wdir):
        if n_clicks is None or url is None:
            raise PreventUpdate

        filenames = get_filenames_from_url(url)
        filenames = [fn for fn in filenames if T.is_ms_file(fn)]

        if len(filenames) == 0:
            return dbc.Alert(f'No MS files found at {url}', color='warning')

        ms_dir = T.get_ms_dirname( wdir )
        fns = []
        n_files = len(filenames)
        for i, fn in enumerate( tqdm( filenames )):
            _url = url+'/'+fn
            logging.info('Downloading', _url)
            fsc.set('progress', int(100*(1+i)/n_files))
            wget.download(_url, out=ms_dir)
        return dbc.Alert(f'{len(fns)} files downloaded.', color='success')


def get_filenames_from_url(url):
    if url.startswith('ftp'):
        return get_files_from_ftp_directory(url)
    if '://' in url:
            url = url.split('://')[1]    
    with urllib3.PoolManager() as http:
        r = http.request('GET', url)
    soup = BeautifulSoup(r.data, 'html')
    files = [A['href'] for A in soup.find_all('a', href=True)]
    return files


def get_files_from_ftp_directory(url):
    url_parts = urlparse(url)
    domain = url_parts.netloc
    path = url_parts.path
    ftp = ftplib.FTP(domain)
    ftp.login()
    ftp.cwd(path)
    filenames = ftp.nlst()
    ftp.quit()
    return filenames




import os
import numpy as np

import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from ms_mint.Mint import Mint
from ms_mint.vis.plotly.plotly_tools import plot_heatmap

from . import tools as T

_label = 'Heatmap'


heatmap_options = [
    { 'label': 'Normalized by biomarker', 'value': 'normed_by_cols'},
    { 'label': 'Cluster', 'value': 'clustered'},
    { 'label': 'Dendrogram', 'value': 'add_dendrogram'},
    { 'label': 'Transposed', 'value': 'transposed'},
    { 'label': 'Correlation', 'value': 'correlation'},
    { 'label': 'Show in new tab', 'value': 'call_show'},
    { 'label': 'log1p', 'value': 'log1p'},
]


_layout = html.Div([
    html.H3('Heatmap'),
    html.Button('Update', id='heatmap-update'),
    dcc.Dropdown(id='heatmap-options', value=['normed_by_cols'],
        options=heatmap_options, multi=True),
    dcc.Loading( 
        dcc.Graph(id='heatmap-figure', 
                  style={'marginTop': '50px'}) ),
])


def layout():
    return _layout


def callbacks(app, fsc, cache):

    @app.callback(
        Output('heatmap-controls', 'children'),
        Input('ana-secondary-tab', 'value'),
        State('wdir', 'children')
    )
    def heat_controls(tab, wdir):
        if tab != _label:
            raise PreventUpdate
        return _layout
        

    @app.callback(
    Output('heatmap-figure', 'figure'),
    Input('heatmap-update', 'n_clicks'),
    State('ana-file-types', 'value'),
    State('ana-peak-labels-include', 'value'),
    State('ana-peak-labels-exclude', 'value'),
    State('ana-ms-order', 'value'),
    State('heatmap-options', 'value'),
    State('wdir', 'children')
    )
    def heat_heatmap(n_clicks, file_types, include_labels, exclude_labels, 
            ms_order, options, wdir):
        mint = Mint()

        df = T.get_complete_results( wdir, include_labels=include_labels, 
                exclude_labels=exclude_labels, file_types=file_types )

        if len(df) == 0: return 'No results yet. First run MINT.'

        mint.results = df

        var_name = 'peak_max'
        data = mint.crosstab(var_name)
        data.index = [ T.Basename(i) for i in data.index ]

        if ms_order is not None and len(ms_order)>0:
            df = df.sort_values(ms_order)
            ms_files = df['MS-file'].drop_duplicates()
            data = data.loc[ms_files]

        data.fillna(0, inplace=True)

        name = var_name
        if 'log1p' in options:
            data = data.apply(np.log1p)
            name = f'log( {var_name}+1 )'

        fig = plot_heatmap(data, 
            normed_by_cols='normed_by_cols' in options, 
            transposed='transposed' in options, 
            clustered='clustered' in options,
            add_dendrogram='add_dendrogram' in options,
            correlation='correlation' in options, 
            call_show='call_show' in options,
            name=name)

        return fig






import numpy as np

import colorlover as cl
import pandas as pd
import plotly_express as px

import plotly.graph_objects as go
import plotly.figure_factory as ff

from collections.abc import Iterable
from os.path import basename
from plotly.subplots import make_subplots

import plotly.io as pio


def set_template():
    pio.templates["draft"] = go.layout.Template(
        layout=dict(font={'size': 10}),
    )

    pio.templates.default = "draft"

set_template()

def plot_peak_shapes(mint_results, n_cols=1, biomarkers=None, peak_labels=None, legend=True, 
                     verbose=False, legend_orientation='v', call_show=False):
    '''
    Returns a plotly multiplost of all peak_shapes in mint.results
    grouped by peak_label.
    '''
    mint_results = mint_results.copy()
    mint_results.ms_file = [basename(i) for i in mint_results.ms_file]
    
    res = mint_results[mint_results.peak_area > 0]

    files = list(res.ms_file.drop_duplicates())
    labels = list(mint_results.peak_label.drop_duplicates())
    
    res = res.set_index(['peak_label', 'ms_file'])
    
    if peak_labels is None:
        peak_labels = []
        #peak_labels = mint_results.groupby('peak_label').mean().peak_max.sort_values(ascending=False).index.astype(str) 

    if biomarkers is None:
        biomarkers = peak_labels

    if biomarkers is None:
        biomarkers = []
        #biomarkers = mint_results.groupby('peak_label').mean().peak_max.sort_values(ascending=False).index.astype(str) 

    if isinstance(biomarkers, str):
        biomarkers = [biomarkers]

    if len(biomarkers) != 0:
        labels = [str(i) for i in biomarkers]
    
    # Calculate neccessary number of rows
    n_rows = len(labels)//n_cols
    if n_rows*n_cols < len(labels):
        n_rows += 1
    
    if verbose:
        print(n_rows, n_cols)
        print('ms_files:', files)
        print('peak_labels:', labels)
        print('Data:', res)
    
    fig = make_subplots(rows=max(1, n_rows), 
                        cols=max(1, n_cols), 
                        subplot_titles=labels)
    if len(files) < 13:
        colors = cl.scales['12']['qual']['Paired']
    else:
        colors = cl.interp( cl.scales['12']['qual']['Paired'], len(files))

    # Create sub-plots
    for label_i, label in enumerate(labels):
        for file_i, file in enumerate(files):
            try:
                x, y = res.loc[(label, file), ['peak_shape_rt', 'peak_shape_int']]
            except:
                continue
            if not isinstance(x,  Iterable):
                continue
            
            if isinstance(x, str):
                x = x.split(',')
                y = y.split(',')
                
            ndx_r = (label_i // n_cols)+1
            ndx_c = label_i % n_cols + 1
                        
            if len(x) == 1:
                mode='markers'
            else:
                mode='lines'
            
            fig.add_trace(
                go.Scatter(
                        x=x, 
                        y=y,
                        name=basename(file),
                        mode=mode,
                        legendgroup=file_i,
                        showlegend=(label_i == 0),  
                        marker_color=colors[file_i],
                        text=file),
                row=ndx_r,
                col=ndx_c,
            )

            fig.update_xaxes(title_text="Scan Time", row=ndx_r, col=ndx_c)
            fig.update_yaxes(title_text="Intensity", row=ndx_r, col=ndx_c)

    # Layout
    if legend:
        fig.update_layout(legend_orientation=legend_orientation)
    
    fig.update_layout(showlegend=legend)
    fig.update_layout(height=400*n_rows, title_text="Peak Shapes")

    if call_show: 
        fig.show(config={'displaylogo': False})
    else:
        return fig


def plot_peak_shapes_3d(mint_results, peak_label=None, legend=True, 
                        legend_orientation='v', call_show=False, verbose=False):
    '''
    Returns a plotly 3D plot of all peak_shapes in mint.results
    where mint.results.peak_label == peak_label.
    '''

    mint_results = mint_results.copy()
    mint_results.ms_file = [basename(i) for i in mint_results.ms_file]

    data = mint_results[mint_results.peak_label == peak_label]
    files = list( data.ms_file.drop_duplicates() )

    grps = data.groupby('ms_file')

    # Peak labels are supposed to be strings
    # Sometimes they are converted to int though
   
    samples = []
    for i, fn in enumerate(files):
        grp = grps.get_group(fn)
        try:
            x, y, peak_max = grp[['peak_shape_rt', 'peak_shape_int', 'peak_max']].values[0]
        except:
            continue
            
        if isinstance(x, str):
            x = x.split(',')
            y = y.split(',')
        sample = pd.DataFrame({'Scan Time': x, 'Intensity': y})
        sample['peak_max'] = peak_max
        sample['ms_file'] = basename(fn)
        samples.append(sample)
    
    if len(samples) == 0:
        return None
    
    samples = pd.concat(samples)
    
    fig = px.line_3d(samples, x='Scan Time', y='peak_max' , z='Intensity', color='ms_file')
    fig.update_layout({'height': 1000, 'width': 1000})
    
    # Layout
    if legend:
        fig.update_layout(legend_orientation=legend_orientation)
    
    fig.update_layout(showlegend=legend)
        
    fig.update_layout({'title': peak_label, 'title_x': 0.5})
    
    if call_show: 
        fig.show(config={'displaylogo': False})
    else:
        return fig



def plot_heatmap(df, normed_by_cols=False, transposed=False, clustered=False,
                 add_dendrogram=False, name='', x_tick_colors=None,
                 correlation=False, call_show=False, verbose=False):


    max_is_not_zero = df.max(axis=1) != 0
    non_zero_labels = max_is_not_zero[max_is_not_zero].index
    df = df.loc[non_zero_labels]

    plot_type = 'Heatmap'
    colorscale = 'Bluered'
    plot_attributes = []
    
    if normed_by_cols:
        df = df.divide(df.max()).fillna(0)
        plot_attributes.append('normalized')

    if transposed:
        df = df.T
        
    if correlation:
        plot_type = 'Correlation'
        df = df.corr()
        colorscale = [[0.0, "rgb(165,0,38)"],
                [0.1111111111111111, "rgb(215,48,39)"],
                [0.2222222222222222, "rgb(244,109,67)"],
                [0.3333333333333333, "rgb(253,174,97)"],
                [0.4444444444444444, "rgb(254,224,144)"],
                [0.5555555555555556, "rgb(224,243,248)"],
                [0.6666666666666666, "rgb(171,217,233)"],
                [0.7777777777777778, "rgb(116,173,209)"],
                [0.8888888888888888, "rgb(69,117,180)"],
                [1.0, "rgb(49,54,149)"]]
    else:
        plot_type = 'Heatmap'
        
    if clustered:
        dendro_side = ff.create_dendrogram(df, orientation='right', labels=df.index.to_list(),
                                           color_threshold=0, colorscale=['black']*8)
        dendro_leaves = dendro_side['layout']['yaxis']['ticktext']
        df = df.loc[dendro_leaves,:]
        if correlation:
            df = df[df.index]        

    x = df.columns
    if clustered:
        y = dendro_leaves
    else:
        y = df.index.to_list()
    z = df.values

    heatmap = go.Heatmap(x=x, y=y, z=z, colorscale=colorscale)
    
    title = f'{plot_type} of {",".join(plot_attributes)} {name}'

    # Figure without side-dendrogram
    if (not add_dendrogram) or (not clustered):
        fig = go.Figure(heatmap)
        fig.update_layout(
            {'title_x': 0.5},
            title={'text': title},
            yaxis={'title': '', 
                   'tickmode': 'array', 
                   'automargin': True}) 
    
        fig.update_layout({'height':800, 
                           'hovermode': 'closest'})
        
    else:  # Figure with side-dendrogram
        fig = go.Figure()
        
        for i in range(len(dendro_side['data'])):
            dendro_side['data'][i]['xaxis'] = 'x2'

        for data in dendro_side['data']:
            fig.add_trace(data)
            
        y_labels = heatmap['y']
        heatmap['y'] = dendro_side['layout']['yaxis']['tickvals']
        
        fig.add_trace(heatmap)     

        fig.update_layout(
                {'height': 800,
                 'showlegend':False,
                 'hovermode': 'closest',
                 'paper_bgcolor': 'white',
                 'plot_bgcolor': 'white',
                 'title_x': 0.5
                },
                title={'text': title},
                
                # X-axis of main figure
                xaxis={'domain': [.11, 1],        
                       'mirror': False,
                       'showgrid': False,
                       'showline': False,
                       'zeroline': False,
                       'showticklabels': True,
                       'ticks':""
                      },
                # X-axis of side-dendrogram
                xaxis2={'domain': [0, .1],  
                        'mirror': False,
                        'showgrid': True,
                        'showline': False,
                        'zeroline': False,
                        'showticklabels': False,
                        'ticks':""
                       },
                # Y-axis of main figure
                yaxis={'domain': [0, 1],
                       'mirror': False,
                       'showgrid': False,
                       'showline': False,
                       'zeroline': False,
                       'showticklabels': False,
                      })

        fig['layout']['yaxis']['ticktext'] = np.asarray(y_labels)
        fig['layout']['yaxis']['tickvals'] = np.asarray(dendro_side['layout']['yaxis']['tickvals'])

    fig.update_layout(
        margin=dict( l=50, r=10, b=200, t=50, pad=0 ),
        hovermode='closest')

    if call_show: 
        fig.show(config={'displaylogo': False})
    else:
        return fig


import os

import dash_html_components as html
import dash_bootstrap_components as dbc
from dash.exceptions import PreventUpdate
from dash_extensions.snippets import send_file

from dash.dependencies import Input, Output, State

from ms_mint.Mint import Mint

from . import tools as T

_label = 'Processing'

_layout = html.Div([
    html.H3('Run MINT'),
    html.Button('Run MINT',         id='run-mint'),
    html.Button('Download results', id='res-download'),
    html.Button('Delete results',   id='res-delete', style={'float': 'right'}),
])

_outputs = html.Div(id='run-outputs', children=[
    html.Div(id={'index': 'run-mint-output', 'type': 'output'}, style={'visibility': 'hidden'}),
])

def layout():
    return _layout


def callbacks(app, fsc, cache):

    @app.callback(
        Output('res-delete-output', 'children'),
        Input('res-delete', 'n_clicks'),
        State('wdir', 'children')
    )
    def heat_delete(n_clicks, wdir):
        if n_clicks is None:
            raise PreventUpdate
        os.remove( T.get_results_fn(wdir) )
        return dbc.Alert('Results file deleted.')


    @app.callback([
        Output('res-download-data', 'data'),
        Input('res-download', 'n_clicks'),
        State('wdir', 'children')
    ])
    def update_link(n_clicks, wdir):
        if n_clicks is None:
            raise PreventUpdate
        fn = T.get_results_fn(wdir)
        workspace = os.path.basename( wdir )
        return [send_file(fn, filename=f'{T.today()}-MINT-results_{workspace}.csv')]


    @app.callback(
        Output({'index': 'run-mint-output', 'type': 'output'}, 'children'),
        Input('run-mint', 'n_clicks'),
        State('wdir', 'children')
    )
    def run_mint(n_clicks, wdir):
        if n_clicks is None:
            raise PreventUpdate

        def set_progress(x):
            fsc.set('progress', x)

        mint = Mint(verbose=True, progress_callback=set_progress)
        mint.peaklist_files = T.get_peaklist_fn( wdir )
        mint.ms_files = T.get_ms_fns( wdir )
        mint.run()
        mint.export( T.get_results_fn(wdir) )
        return dbc.Alert('Finished running MINT', color='success')



from ms_mint.processing import process_ms1_file
from ms_mint.peaklists import read_peaklists
from ms_mint.standards import MINT_RESULTS_COLUMNS

def check_result(result, peaklist):
    assert list(result.columns) == MINT_RESULTS_COLUMNS, list(result.columns)
    assert (result.peak_label == peaklist.peak_label).all(), result.peak_label
    return True

def test__process_ms1_file_mzxml():
    peaklist = read_peaklists('tests/data/peaklist_v1.csv')
    result = process_ms1_file('tests/data/test.mzXML', peaklist=peaklist)
    assert check_result(result, peaklist)

def test__process_ms1_file_mzml():
    peaklist = read_peaklists('tests/data/peaklist_v1.csv')
    result = process_ms1_file('tests/data/test.mzML', peaklist=peaklist)
    assert check_result(result, peaklist)

def test__process_ms1_file_broken_mzml():
    peaklist = read_peaklists('tests/data/peaklist_v1.csv')
    result = process_ms1_file('tests/data/test.mzML', peaklist=peaklist)
    assert check_result(result, peaklist)


import time
from multiprocessing import Pool, Manager, cpu_count



class ParallelExecutor():
    def __init__(self, function, args, processes=None, progress_callback=None):
        self.args = args
        self.function = function
        self._progress = 0
        self.progress_callback = progress_callback

        if processes is None:
            self.processes = cpu_count()
        else:
            self.processes = processes

    def run(self):
        pool = Pool(processes=self.processes)
        m = Manager()
        q = m.Queue()
        args = self.args

        results = pool.map_async(self.function, args)

        while True:
            if results.ready():
                break
            else:
                size = q.qsize()
                progress = int(100*(size+1)/len(args))
                self.progress = progress
                time.sleep(1)

        self.progress = 100

        pool.close()
        pool.join()
        return results.get()

    @property
    def progress(self):
        return self.progress
    
    @progress.setter
    def progress(self, value):
        self._progress = value
        if self.progress_callback is not None:
            self.progress_callback(value)


# ms_mint/standards.py

import os

MINT_ROOT = os.path.dirname(__file__)

PEAKLIST_COLUMNS = ['peak_label', 'mz_mean', 'mz_width', 
                    'rt', 'rt_min', 'rt_max', 'intensity_threshold', 
                    'peaklist_name']

RESULTS_COLUMNS = [
    'peak_area', 
    'peak_n_datapoints', 'peak_max', 
    'peak_rt_of_max',
    'peak_min', 'peak_median',
    'peak_mean', 'peak_delta_int',
    'peak_shape_rt', 'peak_shape_int', 
    'peak_mass_diff_25pc', 'peak_mass_diff_50pc', 
    'peak_mass_diff_75pc',
    'peak_score']

MINT_RESULTS_COLUMNS = (['ms_file']
                        + PEAKLIST_COLUMNS
                        + RESULTS_COLUMNS
                        + ['total_intensity', 'ms_path', 'ms_file_size']
                       )

DEPRECATED_LABELS = {'peakLabel': 'peak_label',
                  #   'compoundId': 'peak_label',
                     'compound': 'peak_label',
                     'peakMz': 'mz_mean',
                     'medRt': 'rt',
                     'medMz': 'mz_mean',
                     'peakMzWidth[ppm]': 'mz_width',
                     'rtmin': 'rt_min',
                     'rtmax': 'rt_max',
                     'peaklist': 'peaklist_name'}

M_PROTON = 1.00782503223



import os
import numpy as np

import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from ms_mint.notebook import Mint
from ms_mint.vis.plotly.plotly_tools import plot_heatmap

from . import tools as T


_label='Hierachical Clustering'

_options = ['Transposed']

options = [{'value': x, 'label': x} for x in _options]

_layout = html.Div([
    html.H3(_label),

    dcc.Input(id='hc-figsize-x', placeholder='Figure size x', value=8, type="number"),
    dcc.Input(id='hc-figsize-y', placeholder='Figure size x', value=8, type="number"),
    dcc.Dropdown(id='hc-options', options=options, value=[]),

    html.Button('Update', id='hc-update'),

    dcc.Loading( 
        html.Div(id='hc-figures',
         style={'margin': 'auto', 
                'text-align': 'center',
                'maxWidth': '100%',
                'minHeight': '300px'}) )

])

_ouptuts = html.Div([])

def layout():
    return _layout


def callbacks(app, fsc, cache):
    
    @app.callback(
        Output('hc-figures', 'children'),
        Input('hc-update', 'n_clicks'),
        State('hc-figsize-x', 'value'),
        State('hc-figsize-y', 'value'),
        State('hc-options', 'value'),
        State('wdir', 'children')
    )
    def create_figure(n_clicks, fig_size_x, fig_size_y, options, wdir):
        if n_clicks is None: raise PreventUpdate

        mint = Mint()

        if fig_size_x is None: fig_size_x = 8
        if fig_size_y is None: fig_size_y = 8

        fig_size_x = min(float(fig_size_x), 100)
        fig_size_y = min(float(fig_size_y), 100)

        print(fig_size_x, fig_size_y)

        mint.results = T.get_results( wdir )

        mint.plot_clustering(figsize=(fig_size_x, fig_size_y), 
            transpose='Transposed' in options)

        src = T.fig_to_src()
        print('HC figure created.')
        return html.Img(src=src, style={'maxWidth': '80%'})


import os
import shutil
import logging 

from tqdm import tqdm

import numpy as np

import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt
import seaborn as sns  

import dash
import dash_bootstrap_components as dbc
import dash_html_components as html
import dash_core_components as dcc

from dash.exceptions import PreventUpdate

from dash.dependencies import Input, Output, State, ALL

import plotly.graph_objects as go

from ms_mint.peak_optimization.RetentionTimeOptimizer import RetentionTimeOptimizer as RTOpt

from ..Mint import Mint

from . import tools as T



info_txt = '''
Creating chromatograms from mzXML/mzML files can last 
a long time the first time. Try converting your files to 
_feather_ format first.'
'''

_label = 'Peak Optimization'

_layout = html.Div([
    html.H3('Peak Optimization'),
    html.H4('File selection'),
    dcc.Dropdown(id='pko-ms-selection',
        options=[
            {'label': 'Use selected files from metadata table (PeakOpt)', 'value': 'peakopt'},
            {'label': 'Use all files (may take a long time)', 'value': 'all'}], 
        value='peakopt',
        clearable=False),

    dcc.Markdown('---'),
    html.H4('Process all peaks'),

    dbc.Row([
        html.Div([
                html.Button('Find largest peaks for all', id='pko-find-largest-peak-for-all',  style={'width': '100%'}),
                html.Label('Margin:'),
                dcc.Slider(id='pko-margin', min=0.1, max=20, step=0.05, value=0.3),
                html.Label('HALLO', id='pko-margin-display'),
            ], style={'width': '45%', 'display': 'inline-block', 'margin': 'auto'}
        ),

        html.Div([
                html.Button('Remove low intensity peaks', id='pko-remove-low-intensity', style={'width': '100%'}),
                html.Label('Threshold:'), 
                dcc.Input(id='pko-threshold', value='1e4'),
            ], style={'width': '45%', 'display': 'inline-block', 'margin': 'auto'}
        ),
    ]),
        
    
    dcc.Markdown('---'),
    html.H4('Peak previews'),

    html.Button('Update peak previews', id='pko-peak-preview'),
    html.Button('Regenerate all figures', id='pko-peak-preview-from-scratch'),

    html.Div(id='pko-peak-preview-images', 
        style={"maxHeight": "600px", "overflowY": "scroll", 'padding': 'auto'}),
    dcc.Markdown('---'),
    html.Div(id='pko-controls'),
    dcc.Dropdown(id='pko-dropdown', options=[], value=None),
    dbc.Progress(id="pko-progress-bar", value=0, 
        style={'marginBottom': '20px', 'width': '100%'}),
    dcc.Loading( dcc.Graph('pko-figure') ),
    dcc.Checklist(id='pko-figure-options', 
                  options=[{'value': 'log','label': 'Logarithmic y-scale'}], 
                  value=[]),

    html.Button('Set RT to current view', id='pko-set-rt'),
    html.Button('Find largest peak', id='pko-find-largest-peak'),
    html.Button('Confirm retention time', id='pko-confirm-rt'),
    html.Button('Remove Peak', id='pko-delete', style={'float': 'right'}),
    
    html.Div(id='pko-image-clicked'),
    html.Div([
        html.Button('<< Previous', id='pko-prev'), 
        html.Button('Suggest', id='pko-suggest-next'),
        html.Button('Next >>', id='pko-next')],
            style={'text-align': 'center', 'margin': 'auto', 'marginTop': '10%'}),

    
])

pko_layout_no_data = html.Div([
    dcc.Markdown('''### No peaklist found.
    You did not generate a peaklist yet.
    ''')
])

_outputs = html.Div(id='pko-outputs', 
    children=[
        html.Div(id={'index': 'pko-set-rt-output', 'type': 'output'}),
        html.Div(id={'index': 'pko-confirm-rt-output', 'type': 'output'}),    
        html.Div(id={'index': 'pko-find-largest-peak-for-all-output', 'type': 'output'}),
        html.Div(id={'index': 'pko-find-largest-peak-output', 'type': 'output'}),
        html.Div(id={'index': 'pko-delete-output', 'type': 'output'}),
        html.Div(id={'index': 'pko-remove-low-intensity-output', 'type': 'output'}),
    ]
)

def layout():
    return _layout 


def callbacks(app, fsc, cache):

    @app.callback(
    Output('pko-dropdown', 'options'),
    Input('tab', 'value'),
    Input({'index': 'pko-delete-output', 'type': 'output'}, 'children'),
    State('wdir', 'children'),
    State('pko-dropdown', 'options'),
    )
    def pko_controls(tab, peak_deleted, wdir, old_options):
        if tab != _label:
            raise PreventUpdate
        peaklist = T.get_peaklist( wdir )
        if peaklist is None:
            logging.warning('Peaklist is None')
            raise PreventUpdate
        options = [{'label':label, 'value': i} for i, label in enumerate(peaklist.index)]
        if options == old_options:
            raise PreventUpdate
        return options


    @app.callback(
    Output('pko-figure', 'figure'),
    Input('pko-dropdown', 'value'),
    Input('pko-figure-options', 'value'),
    Input({'index': 'pko-set-rt-output', 'type': 'output'}, 'children'),
    Input('pko-dropdown', 'options'),
    Input({'index': 'pko-find-largest-peak-for-all-output', 'type': 'output'}, 'children'),
    Input({'index': 'pko-find-largest-peak-output', 'type': 'output'}, 'children'),
    Input({'index': 'pko-confirm-rt-output', 'type': 'output'}, 'children'),
    State('pko-ms-selection', 'value'),
    State('pko-margin', 'value'),
    State('wdir', 'children'),
    #State('pko-figure', 'figure')
    )
    def pko_figure(peak_label_ndx, options, n_clicks, options_changed, 
                   find_largest_peak, find_largest_peak_single, rt_set,
                   ms_selection, margin, wdir,
                   ):
        fig = None
        if peak_label_ndx is None:
            raise PreventUpdate
        peaklist = T.get_peaklist( wdir ).reset_index()
        if ms_selection == 'peakopt':
            ms_files = T.get_ms_fns_for_peakopt(wdir)
        elif ms_selection == 'all':
            ms_files = T.get_ms_fns(wdir)

        cols = ['mz_mean', 'mz_width', 'rt', 
                'rt_min', 'rt_max', 'peak_label']

        peak_label_ndx = peak_label_ndx % len(peaklist)
        mz_mean, mz_width, rt, rt_min, rt_max, label = \
            peaklist.loc[peak_label_ndx, cols]

        if rt is np.isnan(rt):
            if (not np.isnan(rt_min)) and (not np.isnan(rt_max)):
                rt = np.mean([rt_min, rt_max])
        else:
            if (rt_min is None) or np.isnan(rt_min): rt_min = max(0, rt-margin)
            if (rt_max is None) or np.isnan(rt_max): rt_max = rt+margin

        if True or fig is None:
            fig = go.Figure()
            fig.layout.hovermode = 'closest'
            fig.layout.xaxis.range=[rt_min, rt_max]

            fig.update_layout(
                yaxis_title="MS-Intensity",
                xaxis_title="Retention Time [min]",                
                xaxis=dict( 
                    rangeslider=dict( 
                        visible=True
                    )
                )
            )
            fig.update_layout(title=label)
            if 'log' in options: fig.update_yaxes(type="log")

        if not np.isnan(rt):
            fig.add_vline(rt)

        if (not np.isnan(rt_min)) and (not np.isnan(rt_max)):
            fig.add_vrect(x0=rt_min, x1=rt_max, line_width=0, fillcolor="green", opacity=0.1)

        n_files = len(ms_files)
        for i, fn in tqdm(enumerate(ms_files), total=n_files, desc='PKO-figure'):
            fsc.set('progress', int(100*(i+1)/n_files))

            name = os.path.basename(fn)
            name, _ = os.path.splitext(name)
            chrom = T.get_chromatogram(fn, mz_mean, mz_width, wdir)
            fig.add_trace(
                go.Scatter(x=chrom['retentionTime'], 
                           y=chrom['intensity array'], 
                        name=name)
            )
            fig.update_layout(showlegend=False)
            fig.update_layout(hoverlabel = dict(namelength=-1))
        return fig


    @app.callback(
        Output('pko-progress-bar', 'value'),
        Input('pko-dropdown', 'value'),
        State('pko-dropdown', 'options'))
    def set_progress(value, options):
        if (value is None) or (options is None):
            raise PreventUpdate
        progress = int( 100 * (value+1) / len(options))
        return progress


    @app.callback(
    Output({'index': 'pko-find-largest-peak-for-all-output', 'type': 'output'}, 'children'),
    Input('pko-find-largest-peak-for-all', 'n_clicks'),
    State('pko-ms-selection', 'value'),
    State('pko-margin', 'value'),
    State('wdir', 'children'))
    def pko_find_largest_peak(n_clicks, ms_selection, margin, wdir):
        if n_clicks is None:
            raise PreventUpdate
        peaklist = T.get_peaklist( wdir )

        if ms_selection == 'peakopt':
            ms_files = T.get_ms_fns_for_peakopt(wdir)
        elif ms_selection == 'all':
            ms_files = T.get_ms_fns(wdir)

        n_peaks = len(peaklist)
        for i, (peak_label, row) in tqdm( enumerate(peaklist.iterrows()), total=n_peaks ):
            fsc.set('progress', int(100*(1+i)/n_peaks))
            mz_mean, mz_width = row.loc[['mz_mean', 'mz_width']]
            chromatograms = [T.get_chromatogram(fn, mz_mean, mz_width, wdir)\
                .set_index('retentionTime')['intensity array'] for fn in ms_files]
            rt_min, rt_max = max(0, row['rt']-margin), row['rt']+margin
            rt_min, rt_max = RTOpt(rt_min=rt_min, rt_max=rt_max).find_largest_peak(chromatograms)
            peaklist.loc[peak_label, ['rt_min', 'rt_max']] = rt_min, rt_max
        
        T.write_peaklist( peaklist, wdir)
        return dbc.Alert('Peak optimization done.', color='info')

    
    @app.callback(
        Output({'index': 'pko-set-rt-output', 'type': 'output'}, 'children'),
        Input('pko-set-rt', 'n_clicks'),
        State('pko-dropdown', 'value'),
        State('pko-figure', 'figure'),
        State('wdir', 'children')
    )
    def pko_set_rt_min_max(n_clicks, peak_label, fig, wdir):
        if n_clicks is None:
            raise PreventUpdate
        rt_min, rt_max = fig['layout']['xaxis']['range']
        rt_min, rt_max = np.round(rt_min, 4), np.round(rt_max, 4)
        T.update_peaklist(wdir, peak_label, rt_min, rt_max)
        return dbc.Alert(f'Set RT span to ({rt_min},{rt_max})', color='info')
    
    
    @app.callback(
        Output({'index': 'pko-confirm-rt-output', 'type': 'output'}, 'children'),
        Input('pko-confirm-rt', 'n_clicks'),
        State('pko-dropdown', 'value'),
        State('pko-figure', 'figure'),
        State('wdir', 'children')
    )
    def pko_confirm_rt(n_clicks, peak_label, fig, wdir):
        if n_clicks is None:
            raise PreventUpdate
        rt_min, rt_max = fig['layout']['xaxis']['range']
        rt_min, rt_max = np.round(rt_min, 4), np.round(rt_max, 4)
        
        image_label = f'{peak_label}_{rt_min}_{rt_max}'

        _, fn = T.get_figure_fn(kind='peak-preview', wdir=wdir, 
            label=image_label, format='png')
        
        rt = np.mean([rt_min, rt_max])
        T.update_peaklist(wdir, peak_label, rt=rt)
        if os.path.isfile(fn): os.remove(fn)

        return dbc.Alert(f'Set RT span to ({rt_min},{rt_max})', color='info')
    

    @app.callback(
        Output('pko-dropdown', 'value'),
        Input('pko-prev', 'n_clicks'),
        Input('pko-suggest-next', 'n_clicks'),
        Input('pko-next', 'n_clicks'),
        Input('pko-image-clicked', 'children'),
        State('pko-dropdown', 'value'),
        State('pko-dropdown', 'options'),
        State('wdir', 'children')
    )
    def pko_prev_next_suggest(n_prev, n_suggest, n_next, image_clicked, 
            value, options, wdir):
        if n_prev is None\
            and n_next is None \
            and image_clicked is None\
            and n_suggest is None:
                raise PreventUpdate

        prop_id = dash.callback_context.triggered[0]['prop_id']        

        if prop_id.startswith('pko-suggest'):
            peaklist = T.get_peaklist( wdir ).reset_index()
            rt_means = peaklist[['rt_min', 'rt_max']].mean(axis=1)
            peak_label_ndx = np.argmax( (peaklist.rt-rt_means).abs() )
            return peak_label_ndx      

        if prop_id.startswith('pko-image-clicked'):
            for entry in options:
                if entry['label'] == image_clicked:
                    return entry['value']
        elif value is None:
            return 0
        elif prop_id.startswith('pko-prev'):
            return (value - 1) % len(options)
        elif prop_id.startswith('pko-next'):
            return (value + 1) % len(options)


    @app.callback(
    Output('pko-peak-preview-images', 'children'),
    Input('pko-peak-preview', 'n_clicks'),
    Input('pko-peak-preview-from-scratch', 'n_clicks'),
    State('pko-ms-selection', 'value'),
    State('wdir', 'children')
    )
    def peak_preview(n_clicks, from_scratch, #peak_opt, #set_rt,
            ms_selection, wdir):
        if n_clicks is None:
            raise PreventUpdate
        # reset updating after 5 attempts
        n_attempts = fsc.get(f'{wdir}-update-attempt')
        if n_attempts is None:
            n_attempts = 1
        elif n_attempts % 5 :
            fsc.set(f'{wdir}-updating', False)
        
        # increment counter of attempts
        fsc.set(f'{wdir}-update-attempt', n_attempts + 1) 

        if fsc.get(f'{wdir}-updating') is True :
            raise PreventUpdate

        fsc.set(f'{wdir}-updating', True)

        prop_id = dash.callback_context.triggered[0]['prop_id']
        regenerate = prop_id.startswith('pko-peak-preview-from-scratch')
        if regenerate:
            image_path = os.path.join( wdir, 'figures', 'peak-preview')
            if os.path.isdir(image_path):
                shutil.rmtree(image_path)

        if ms_selection == 'peakopt':
            ms_files = T.get_ms_fns_for_peakopt(wdir)
        elif ms_selection == 'all':
            ms_files = T.get_ms_fns(wdir)
        else:
            assert False, ms_selection

        if len(ms_files)==0:
            return dbc.Alert('No files selected for peak optimization in Metadata tab. Please, select some files in column "PeakOpt".', color='warning')
        else:
            logging.info(f'Using {len(ms_files)} files for peak preview. ({ms_selection})')

        peaklist = T.get_peaklist(wdir)

        file_colors = T.file_colors( wdir )

        n_total = len(peaklist)
        
        sns.set_context('paper')
        images = []
        for i, (peak_label, row) in tqdm( enumerate( peaklist.iterrows() ), total=n_total):
            fsc.set('progress', int(100*(i+1) / n_total ))
            mz_mean, mz_width, rt, rt_min, rt_max = \
                row[['mz_mean', 'mz_width', 'rt', 'rt_min', 'rt_max']]

            if np.isnan(rt_min) or rt_min is None: rt_min = 0
            if np.isnan(rt_max) or rt_max is None: rt_max = 15

            image_label = f'{peak_label}_{rt_min}_{rt_max}'

            _, fn = T.get_figure_fn(kind='peak-preview', wdir=wdir, 
                label=image_label, format='png')

            if not os.path.isfile( fn ) or regenerate:
                create_preview_peakshape(ms_files, mz_mean, mz_width, rt, 
                    rt_min, rt_max, image_label, wdir, title=peak_label, colors=file_colors)

            if os.path.isfile(fn):
                src = T.png_fn_to_src(fn)
            else:
                src = None

            _id = {'index': peak_label, 'type': 'image'}
            image_id = f'image-{i}'
            images.append(
                html.A(id=_id, 
                children=html.Img(src=src, height=300, id=image_id, 
                    style={'margin': '10px'}))
            )
            images.append( 
                dbc.Tooltip(peak_label, target=image_id, style={'font-size': '50'})
            )
        fsc.set(f'{wdir}-updating', False)
        return images


    @app.callback(
        Output('pko-image-clicked', 'children'),
        # Input needs brakets to make prevent_initital_call work
        [Input({'type': 'image', 'index': ALL}, 'n_clicks')],
        prevent_initial_call=True
        )
    def pko_image_clicked(ndx):
        if ndx is None or len(ndx)==0: raise PreventUpdate
        ctx = dash.callback_context
        clicked = ctx.triggered[0]['prop_id']
        clicked = clicked.replace('{"index":"', '')
        clicked = clicked.split('","type":')[0].replace('\\', '')
        if len( dash.callback_context.triggered) > 1: raise PreventUpdate
        return clicked


    @app.callback(
        Output({'index': 'pko-delete-output', 'type': 'output'}, 'children'),
        Input('pko-delete', 'n_clicks'),
        State('pko-dropdown', 'value'),
        State('wdir', 'children')
    )
    def plk_delete(n_clicks, peak_ndx, wdir):
        if n_clicks is None:
            raise PreventUpdate
        peaklist = T.get_peaklist( wdir ).reset_index()
        peak_label = peaklist.loc[peak_ndx, 'peak_label']
        peaklist = peaklist.drop( peak_ndx, axis=0 )
        T.write_peaklist(peaklist, wdir)
        return dbc.Alert(f'{peak_label} removed from peaklist.', color='info')


    @app.callback(
        Output({'index': 'pko-find-largest-peak-output', 'type': 'output'}, 'children'),
        Input('pko-find-largest-peak', 'n_clicks'),
        State('pko-dropdown', 'value'),
        State('pko-ms-selection', 'value'),
        State('pko-margin', 'value'),
        State('wdir', 'children')
    )
    def find_largest_peak(n_clicks, peak_label_ndx, ms_selection, margin, wdir):
        if n_clicks is None: raise PreventUpdate
        if peak_label_ndx is None: raise PreventUpdate
        peaklist = T.get_peaklist( wdir )
        if ms_selection == 'peakopt':
            ms_files = T.get_ms_fns_for_peakopt(wdir)
        elif ms_selection == 'all':
            ms_files = T.get_ms_fns(wdir)           
        row = peaklist.iloc[peak_label_ndx]
        mz_mean, mz_width = row.loc[['mz_mean', 'mz_width']]
        chromatograms = [T.get_chromatogram(fn, mz_mean, mz_width, wdir)\
            .set_index('retentionTime')['intensity array'] for fn in ms_files]
        rt_min, rt_max = RTOpt(rt=row['rt'], rt_min=row['rt_min'], rt_max=['rt_max'], rt_margin=margin)\
                            .find_largest_peak(chromatograms)
        peaklist = peaklist.reset_index()
        peaklist.loc[peak_label_ndx, ['rt_min', 'rt_max']] = rt_min, rt_max
        peak_label = peaklist.loc[peak_label_ndx, 'peak_label']
        peaklist.to_csv( T.get_peaklist_fn( wdir ), index=False )   
        return dbc.Alert(f'Set rt_min, rt_max for {peak_label}: {rt_min}, {rt_max}.', color='info')


    @app.callback(
        Output({'index': 'pko-remove-low-intensity-output', 'type': 'output'}, 'children'),
        Input('pko-remove-low-intensity', 'n_clicks'),
        State('pko-ms-selection', 'value'),
        State('pko-threshold', 'value'),
        State('wdir', 'children'),
    )
    def remove_low_intensity_peaks(n_clicks, ms_selection, threshold, wdir):
        if n_clicks is None: raise PreventUpdate
        logging.info('Remove low intensity peaks.')
        peaklist = T.get_peaklist( wdir )
        if ms_selection == 'peakopt':
            ms_files = T.get_ms_fns_for_peakopt( wdir )
        elif ms_selection == 'all':
            ms_files = T.get_ms_fns( wdir )

        def set_progress(x):
            fsc.set('progress', x)

        mint = Mint(verbose=True, progress_callback=set_progress)

        tmp_peaklist = peaklist.reset_index().copy()

        tmp_peaklist['rt_min'] = tmp_peaklist.rt_min.fillna(0)
        tmp_peaklist['rt_max'] = tmp_peaklist.rt_max.fillna(100)
        tmp_peaklist['intensity_threshold'] = float(threshold)

        mint.ms_files = ms_files
        mint.peaklist = tmp_peaklist
        mint.run()
        peak_labels = mint.results.peak_label.drop_duplicates()
        peaklist = peaklist[peaklist.index.isin(peak_labels)]
        T.write_peaklist(peaklist, wdir)
        return dbc.Alert('Low intensity peaks removed.', color='info')

    @app.callback(
        Output('pko-margin-display', 'children'),
        Input('pko-margin', 'value'))
    def display_slider_value(value):
        return 'Value: {} [min]'.format(value)


def create_preview_peakshape(ms_files, mz_mean, mz_width, rt, 
        rt_min, rt_max, image_label, wdir, title, colors):
    plt.figure(figsize=(4,2.5), dpi=30)
    y_max = 0 
    for fn in ms_files:
        color = colors[ T.filename_to_label(fn) ]
        fn_chro = T.get_chromatogram(fn, mz_mean, mz_width, wdir)
        fn_chro = fn_chro[(rt_min < fn_chro['retentionTime']) &
                             (fn_chro['retentionTime'] < rt_max)   ]
        plt.plot(fn_chro['retentionTime'], fn_chro['intensity array'], lw=1, color=color)
        y_max = max(y_max, fn_chro['intensity array'].max())
    if (not np.isnan(rt)) \
        and not (np.isnan(rt_max)) \
        and not (np.isnan(rt_min)):
        x = max(min(rt, rt_max), rt_min)
        rt_mean = np.mean([rt_min, rt_max])
        color_value = np.abs(rt_mean-rt)
        color = T.float_to_color(color_value, vmin=0, vmax=1, cmap='coolwarm')
        plt.vlines(x, 0, y_max, lw=3, color=color)
    plt.gca().set_title(title[:30], y=1.0, pad=15)
    plt.gca().ticklabel_format(axis='y', style='sci', scilimits=(0,0))
    plt.xlabel('Retention Time [min]')  
    plt.ylabel('MS-Intensity')
    filename = T.savefig(kind='peak-preview', wdir=wdir, label=image_label)
    plt.close()
    return filename 

  


import os
from ms_mint.Mint import Mint

def test__export_to_excel(tmp_path):
    filename = os.path.join(tmp_path, 'output.xlsx')
    mint = Mint(verbose=True)
    mint.files = 'tests/data/test.mzXML'
    mint.run()
    mint.export(filename)
    assert os.path.isfile(filename)



import os
import numpy as np

import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from ms_mint.notebook import Mint
from ms_mint.vis.plotly.plotly_tools import plot_heatmap

import matplotlib as mpl
import seaborn as sns
sns.set_context('paper')


from . import tools as T


_label='Plotting'

_kinds = ['bar', 'violin', 'box', 'count', 'boxen', 'scatter', 'line', 'strip', 'swarm', 'point']

kind_options = [{'value': x, 'label': x.capitalize()} for x in _kinds]

options = [
    {'label': 'Rotate x-ticks', 'value': 'rot-x-ticks'},
    {'label': 'Share x-axis', 'value': 'share-x'},
    {'label': 'Share y-axes', 'value': 'share-y'},
    {'label': 'Scientific notation', 'value': 'sci'},
    {'label': 'Logarithmic x-scale', 'value': 'log-x'},
    {'label': 'Logarithmic y-scale', 'value': 'log-y'},
    {'label': 'High Quality', 'value': 'HQ'},

]

_layout = html.Div([
    html.H3(_label),

    dcc.Input(id='plot-fig-height', placeholder='Figure facet height x', value=2.5, type="number"),
    dcc.Input(id='plot-fig-aspect', placeholder='Figure aspect ratio', value=1, type="number"),
    dcc.Input(id='plot-title', placeholder='Title', value=None),
    dcc.Dropdown(id='plot-kind', options=kind_options, value='bar'),

    dcc.Dropdown(id='plot-x', options=[], value=None, placeholder='X'),
    dcc.Dropdown(id='plot-y', options=[], value='peak_max', placeholder='Y'),
    dcc.Dropdown(id='plot-hue', options=[], value=None, placeholder='Color'),
    dcc.Dropdown(id='plot-col', options=[], value=None, placeholder='Columns'),
    dcc.Dropdown(id='plot-row', options=[], value=None, placeholder='Rows'),
    dcc.Dropdown(id='plot-style', options=[], value=None, placeholder='Style'),
    dcc.Dropdown(id='plot-size', options=[], value=None, placeholder='Size'),

    html.Label('Column wrap:'),
    dcc.Slider(id='plot-col-wrap', step=1, min=0, max=30, value=0),
    dcc.Dropdown(id='plot-options', value=['sci'], options=options, multi=True), 

    html.Button('Update', id='plot-update'),

    dcc.Loading( 
        html.Div(id='plot-figures',
         style={'margin': 'auto',
                'marginTop': '10%',
                'text-align': 'center',
                'maxWidth': '100%',
                'minHeight': '300px'}) )

])

_ouptuts = html.Div([])

def layout():
    return _layout


def callbacks(app, fsc, cache):

    @app.callback(
        Output('plot-x', 'options'),
        Output('plot-y', 'options'),
        Output('plot-col', 'options'),
        Output('plot-row', 'options'),
        Output('plot-hue', 'options'),
        Output('plot-size', 'options'),
        Output('plot-style', 'options'),
        Input('ana-secondary-tab', 'value'),
        State('wdir', 'children')
    )
    def fill_options(tab, wdir):
        if tab != _label: raise PreventUpdate
        results = T.get_complete_results( wdir )
        results = results.dropna(axis=1, how='all')
        cols = results.columns
        options = [{'value': x, 'label': x} for x in cols]
        return [options]*7

    @app.callback(
        Output('plot-figures', 'children'),
        Input('plot-update', 'n_clicks'),
        State('plot-kind', 'value'),
        State('plot-fig-height', 'value'),
        State('plot-fig-aspect', 'value'),
        State('plot-x', 'value'),
        State('plot-y', 'value'),
        State('plot-hue', 'value'),       
        State('plot-col', 'value'),
        State('plot-row', 'value'),
        State('plot-col-wrap', 'value'),
        State('plot-style', 'value'),
        State('plot-size', 'value'),
        State('plot-title', 'value'),
        State('plot-options', 'value'),
        State('wdir', 'children')
    )
    def create_figure(n_clicks, kind, height, aspect, x, y, hue, 
            col, row, col_wrap, style, size, title, options, wdir):

        if n_clicks is None: raise PreventUpdate
        if col_wrap == 0: col_wrap = None
        mint = Mint()
        if col is None and row is None: col_wrap=None
        if height is None: height = 2.5
        if aspect is None: aspect = 1

        height = min(float(height), 5)
        height = max(height, 1)
        aspect = max(.5, float(aspect))
        aspect = min(aspect, 10)
        
        df = T.get_complete_results( wdir )

        n_c, n_r = 1, 1
        if col is not None:
            n_c = len(df[col].drop_duplicates())
        if row is not None:
            n_r = len(df[row].drop_duplicates())
        
        if (n_c is not None) and (col_wrap is not None):
            n_c = n_c // col_wrap
            n_r = n_c % col_wrap
        
        if hue is not None:
            print(hue, df[hue].value_counts(), df[hue].isna().sum())

        print(n_c, n_r)

        if hue is not None and ((x is None) and (y is None)):
            x = hue

        if kind in ['scatter', 'line']:
            plot_func = sns.relplot
            kwargs = dict(facet_kws=dict(
                            sharex='share-x' in options,
                            sharey='share-y' in options,
                            legend_out=True
                          ),
                          style=style,
                          size=size
                     )
        else:
            plot_func = sns.catplot
            kwargs=dict(
                sharex='share-x' in options,
                sharey='share-y' in options,
                facet_kws=dict(legend_out=True)
            )

        g = plot_func(
                x=x, 
                y=y,
                hue=hue, 
                col=col, 
                row=row, 
                col_wrap=col_wrap,
                data=df, 
                kind=kind,
                height=height, 
                aspect=aspect,
                **kwargs
                )

        g.fig.subplots_adjust(top=0.9)
        g.set_titles(col_template="{col_name}", row_template="{row_name}")
        
        if 'log-x' in options:
            g.set(xscale="log")
        if 'log-y' in options:
            g.set(yscale="log")

        notation = 'sci' if 'sci' in options else 'plain'
        for ax in g.axes.flatten():
            try:
                ax.ticklabel_format(style=notation, scilimits=(0,0), axis='x')
            except:
                pass
            try:
                ax.ticklabel_format(style=notation, scilimits=(0,0), axis='y')
            except:
                pass            


        if 'rot-x-ticks' in options:
            g.set_xticklabels(rotation=90)

        if title is not None: g.fig.suptitle(title)


        g.tight_layout(w_pad=0)
   
        print(sns.__version__)
        print(mpl.__version__)
        #g.add_legend()
        try:
            g.legend.set_bbox_to_anchor((1.2, 0.7))
        except:
            pass

        src = T.fig_to_src(dpi=300 if 'HQ' in options else None)

        print('Figure created.')
        return html.Img(src=src, style={'maxWidth': '80%'})



import dash_html_components as html
#import dash_bootstrap_components as dbc
import dash_core_components as dcc
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from . import heatmap
from . import pca
from . import distributions
from . import hierachical_clustering
from . import plotting
from . import tools as T


_modules = [
  heatmap,
  distributions,
  pca,
  hierachical_clustering,
  plotting
]

modules = {module._label: module for module in _modules}

groupby_options = [{'label': 'Batch', 'value': 'Batch'},
                   {'label': 'Label', 'value': 'Label'},
                   {'label': 'Type',  'value': 'Type'},
                   {'label': 'Color', 'value': 'Color'}]


_layout = html.Div([
    dcc.Tabs(id='ana-secondary-tab', value=_modules[0]._label, vertical=False, 
        children=[dcc.Tab(value=key, 
                    label=modules[key]._label,
                    ) for key in modules.keys()
        ]
    ),
    
    dcc.Dropdown(id='ana-file-types', options=[], placeholder='Types of files to include', multi=True),
    dcc.Dropdown(id='ana-peak-labels-include', options=[], placeholder='Include peak_labels', multi=True),
    dcc.Dropdown(id='ana-peak-labels-exclude', options=[], placeholder='Exclude peak_labels', multi=True),    
    dcc.Dropdown(id='ana-ms-order', options=[], placeholder='MS-file sorting', multi=True),
    dcc.Dropdown(id='ana-groupby', options=groupby_options, value=None, placeholder='Group by column'),
    html.Div(id='ana-secondary-tab-content')
])


_label = 'Analysis'

_outputs = None

def layout():
    return _layout


def callbacks(app, fsc, cache):

    for module in _modules:
        func = module.callbacks
        if func is not None:
            func(app=app, fsc=fsc, cache=cache)

    @app.callback(
        Output('ana-secondary-tab-content', 'children'),
        Input('ana-secondary-tab', 'value'),
        State('wdir', 'children')
    )
    def render_content(tab, wdir):
        func = modules[tab].layout
        if func is not None:
            return func()
        else:
            raise PreventUpdate


    @app.callback(
        Output('ana-file-types', 'options'),
        Output('ana-file-types', 'value'),
        Input('tab', 'value'),
        State('wdir', 'children')
    )
    def file_types(tab, wdir):
        if tab != _label:
            raise PreventUpdate
        meta = T.get_metadata( wdir )
        if meta is None:
            raise PreventUpdate
        file_types = meta['Type'].drop_duplicates()
        options = [{'value': i, 'label': i} for i in file_types]
        return options, file_types


    @app.callback(
        Output('ana-ms-order', 'options'),
        Output('ana-groupby', 'options'),
        Input('ana-secondary-tab', 'value'),
        State('wdir', 'children')
    )
    def ms_order_options(tab, wdir):
        cols = T.get_metadata(wdir).dropna(how='all', axis=1).columns.to_list()
        if 'index' in cols: cols.remove('index')
        if 'PeakOpt' in cols: cols.remove('PeakOpt')
        options = [{'value':i, 'label': i} for i in cols]
        return options, options


    @app.callback(
        Output('ana-peak-labels-include', 'options'),
        Output('ana-peak-labels-exclude', 'options'),
        Input('tab', 'value'),
        State('wdir', 'children')
    )
    def peak_labels(tab, wdir):
        if tab != _label:
            raise PreventUpdate
        peaklist = T.get_peaklist( wdir ).reset_index()
        options = [{'value': i, 'label': i} for i in peaklist.peak_label]
        return options, options


import os
import pandas as pd

import dash_bootstrap_components as dbc
import dash_html_components as html
import dash_core_components as dcc

from dash.dependencies import Input, Output, State

from dash.exceptions import PreventUpdate
from dash_tabulator import DashTabulator

from ms_mint import MINT_DATA_PATH
from ms_mint.standards import M_PROTON
from . import tools as T


options = {
           "selectable": True,
           "headerFilterLiveFilterDelay":3000,
           "layout": "fitDataFill",
           "height": "600px",
           }

clearFilterButtonType = {"css": "btn btn-outline-dark", "text":"Clear Filters"}

fn_data = os.path.join(MINT_DATA_PATH, 'ChEBI-Chem.parquet')
fn_groups_data = os.path.join(MINT_DATA_PATH, 'ChEBI-Groups.parquet')

CHEBI_CHEM = pd.read_parquet(fn_data)
CHEBI_GROUPS = pd.read_parquet(fn_groups_data)#.set_index('Group Name')


groups_options = [{'label': 'All', 'value': 'all'}]+[{'label': x.capitalize(), 'value':x} for x in CHEBI_GROUPS.index]

columns = T.gen_tabulator_columns(CHEBI_CHEM.columns, editor=None, col_width='auto')

add_metab_table = html.Div(id='add-metab-table-container', 
    style={'minHeight':  100, 'marginTop': '10%'},
    children=[
        dcc.Dropdown(id='add-metab-ms-mode', placeholder='Select ionization mode',
            options=[{'label': 'Positive', 'value': 'Positive'},
                     {'label': 'Negative', 'value': 'Negative'}], 
            value='Negative'),
        
        dcc.Dropdown(id='add-metab-groups', options=groups_options, multi=True),

        dcc.Loading( 
            DashTabulator(id='add-metab-table',
                columns=columns, 
                options=options,
                clearFilterButtonType=clearFilterButtonType,
            ),
        ),
    html.Button('Add selected metabolites to peaklist', id='add-metab'),
])

_label = 'Add Metabolites'

_layout = html.Div([
    html.H3('Add Metabolites'),
    add_metab_table
    
])

_outputs = html.Div(id='add-metab-outputs', 
    children=[
        html.Div(id={'index': 'add-metab-output', 'type': 'output'})
    ]
)

def layout():
    return _layout


def callbacks(app, fsc, cache):

    @app.callback(
        Output({'index': 'add-metab-output', 'type': 'output'}, 'children'),
        Input('add-metab', 'n_clicks'),
        State('add-metab-table', 'multiRowsClicked'),
        State('add-metab-ms-mode', 'value'),
        State('wdir', 'children')
    )
    def add_metab(n_clicks, rows, ms_mode, wdir):
        if n_clicks is None:
            raise PreventUpdate
        if ms_mode is None:
            return dbc.Alert('Please select ionization mode.', color='warning')

        peaklist = T.get_peaklist( wdir )
        
        for row in rows:
            charge = int( row['Charge'] )
            if (ms_mode=='Negative') and (charge>0):
                continue
            if (ms_mode=='Positive') and (charge<0):
                continue
            if charge == 0:
                if ms_mode == 'Negative':         
                    peaklist.loc[row['ChEBI Name'], 'mz_mean'] = row['Monoisotopic Mass'] - M_PROTON
                elif ms_mode == 'Positive':
                    peaklist.loc[row['ChEBI Name'], 'mz_mean'] = row['Monoisotopic Mass'] #+ M_PROTON
            else:
                peaklist.loc[row['ChEBI Name'], 'mz_mean'] = row['Monoisotopic Mass']

            peaklist.loc[row['ChEBI Name'], 'mz_width'] = 10
            peaklist.loc[row['ChEBI Name'], 'rt'] = -1
            peaklist.loc[row['ChEBI Name'], 'rt_min'] = 0
            peaklist.loc[row['ChEBI Name'], 'rt_max'] = 15
            peaklist.loc[row['ChEBI Name'], 'intensity_threshold'] = 0
            
        T.write_peaklist( peaklist, wdir)
        n_peaks = len(peaklist)
        n_new = len(rows)
        return dbc.Alert(f'{n_new} peaks added, now {n_peaks} peaks defined.', color='info')


    @app.callback(
        Output('add-metab-table', 'data'),
        Input('add-metab-groups', 'value'),
        Input('add-metab-ms-mode', 'value'),
    )
    def generate_table(groups, ms_mode):
        print('Groups to add', groups)

        data = CHEBI_CHEM

        if ms_mode == 'Positive':
            data = data[data.Charge.str.startswith('+') | (data.Charge == '0')]
        elif ms_mode == 'Negative':
            data = data[data.Charge.str.startswith('-') | (data.Charge == '0')]
        
        if groups is None or len(groups) == 0:
            raise PreventUpdate
        elif 'all' in groups:
            return data.to_dict('records')

        ids = CHEBI_GROUPS.loc[groups]\
                .explode('ChEBI IDs')\
                .drop_duplicates()['ChEBI IDs']\
                .values
        
        return data[data['ChEBI ID'].isin(ids)].to_dict('records')


# ms_mint/Mint.py

import os
import numpy as np
import pandas as pd
import time
import logging

from warnings import simplefilter
from pathlib import Path as P

from matplotlib import pyplot as plt
import seaborn as sns

from multiprocessing import Pool, Manager, cpu_count
from scipy.cluster.hierarchy import ClusterWarning
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler, RobustScaler


from .standards import MINT_RESULTS_COLUMNS, PEAKLIST_COLUMNS, DEPRECATED_LABELS
from .processing import process_ms1_files_in_parallel, extract_chromatogram_from_ms1
from .io import export_to_excel, ms_file_to_df
from .peaklists import read_peaklists, check_peaklist, standardize_peaklist
from .peak_detection import OpenMSFFMetabo
from .helpers import is_ms_file, get_ms_files_from_results
from .vis.plotly.plotly_tools import plot_heatmap
from .vis.mpl import plot_peak_shapes, hierarchical_clustering
from .peak_optimization.RetentionTimeOptimizer import RetentionTimeOptimizer

from tqdm import tqdm

import ms_mint

class Mint(object):
    def __init__(self, verbose:bool=False, progress_callback=None):
        self._verbose = verbose
        self._version = ms_mint.__version__
        self._progress_callback = progress_callback
        self.reset()
        if self.verbose:
            print('Mint Version:', self.version , '\n')
        self.peak_detector = OpenMSFFMetabo(progress_callback=progress_callback)

    @property
    def verbose(self):
        return self._verbose
    
    @verbose.setter
    def verbose(self, value:bool):
        self._verbose = value
    
    @property
    def version(self):
        return self._version
    
    def reset(self):
        self._files = []
        self._peaklist_files = []
        self._peaklist = pd.DataFrame(columns=PEAKLIST_COLUMNS)
        self._results = pd.DataFrame({i: [] for i in MINT_RESULTS_COLUMNS})
        self._all_df = None
        self._progress = 0
        self.runtime = None
        self._status = 'waiting'
        self._messages = []

    def optimize_retention_times(self, ms_files=None, peak_labels=None, **kwargs):
        chromatograms = []
        if ms_files is None:
            ms_files = self.ms_files
        if peak_labels is None:
            peak_labels = self.peaklist.peak_label.values
        peaklist = self.peaklist
        n_peaks = len(peaklist)
        for i, (ndx, row) in tqdm( enumerate(peaklist.iterrows()), total=n_peaks ):
            progress = int(100*(i+1)/n_peaks)
            if self.progress_callback is not None: self.progress_callback(progress)
            peak_label = row['peak_label']
            if peak_label not in peak_labels:
                continue
            chromatograms = []
            mz_mean, mz_width, rt, rt_min, rt_max = row[['mz_mean', 'mz_width', 'rt', 'rt_min', 'rt_max']]
            for fn in ms_files:
                df = ms_file_to_df(fn)
                chrom = extract_chromatogram_from_ms1(df, mz_mean=mz_mean, mz_width=mz_width)
                chromatograms.append(chrom)
            rt_min, rt_max = RetentionTimeOptimizer(**kwargs).find_largest_peak(chromatograms)
            self.peaklist.loc[ndx, ['rt_min', 'rt_max']] =  rt_min, rt_max
            

    def clear_peaklist(self):
        self.peaklist = pd.DataFrame(columns=PEAKLIST_COLUMNS)

    def clear_results(self):
        self.results = pd.DataFrame(columns=MINT_RESULTS_COLUMNS)

    def clear_ms_files(self):
        self.ms_files = []

    def run(self, nthreads=None, rt_margin=.5, mode='standard'):
        '''
        Run MINT with set up peaklist and ms-files.
        ----
        Args
            - nthreads: int or None, default = None
                * None: Run with min(n_cpus, c_files) CPUs
                * 1: Run without multiprocessing on one CPU
                * >1: Run with multiprocessing enabled using nthreads threads.
            - mode: str, default = 'standard'
                * 'standard': calculates peak shaped projected to RT dimension
                * 'express': omits calculation of other features, only peak_areas
        '''
        self._status = 'running'
            
        if (self.n_files == 0) or ( len(self.peaklist) == 0):
            return None

        peaklist = self.peaklist
        if 'rt' in peaklist.columns:
            ndx = ((peaklist.rt_min.isna()) & (~peaklist.rt.isna()))
            peaklist.loc[ndx, 'rt_min'] = peaklist.loc[ndx, 'rt'] - rt_margin
            ndx = ((peaklist.rt_max.isna()) & (~peaklist.rt.isna()))
            peaklist.loc[ndx, 'rt_max'] = peaklist.loc[ndx, 'rt'] + rt_margin
            del ndx

        if nthreads is None:
            nthreads = min(cpu_count(), self.n_files)
            
        if self.verbose: print(f'Run MINT with {nthreads} processes:')
        
        start = time.time()
        if nthreads > 1:
            self.run_parallel(nthreads=nthreads, mode=mode)
        else:
            results = []
            for i, filename in enumerate(self.ms_files):
                args = {'filename': filename,
                        'peaklist': self.peaklist,
                        'q': None, 
                        'mode': mode}
                results.append(process_ms1_files_in_parallel(args))
                self.progress = int(100 * (i / self.n_files))
            self.results = pd.concat(results)
            self.progress = 100
        
        end = time.time()
        self.runtime = ( end - start )
        self.runtime_per_file = (self.runtime / self.n_files)
        self.runtime_per_peak = (self.runtime / self.n_files / len(self.peaklist))
        
        if self.verbose: 
            print(f'Total runtime: {self.runtime:.2f}s')
            print(f'Runtime per file: {self.runtime_per_file:.2f}s')
            print(f'Runtime per peak ({len(self.peaklist)}): {self.runtime_per_peak:.2f}s\n')
            print('Results:', self.results )
        self._status = 'done'

    def detect_peaks(self, **kwargs):
        detected = self.peak_detector.fit_transform(self.ms_files, **kwargs)
        if detected is not None:
            self.peaklist = pd.concat([self.peaklist, detected])

    def run_parallel(self, nthreads=1, mode='standard'):
        pool = Pool(processes=nthreads)
        m = Manager()
        q = m.Queue()
        args = []
        for i, filename in enumerate(self.ms_files):
            args.append({'filename': filename,
                         'peaklist': self.peaklist,
                         'queue': q,
                         'mode': mode})
                   
        results = pool.map_async(process_ms1_files_in_parallel, args)
        
        # monitor progress
        while True:
            if results.ready():
                break
            else:
                size = q.qsize()
                self.progress = 100 * size / self.n_files
                time.sleep(1)
    
        self.progress = 100

        pool.close()
        pool.join()
        results = results.get()
        self.results = pd.concat(results)

    @property
    def messages(self):
        return self._messages
             
    @property
    def status(self):
        return self._status

    @property
    def files(self):
        return self._files

    @property
    def ms_files(self):
        return self._files
    
    @files.setter
    def files(self, list_of_files):
        print('Mint.files is deprecated, please use Mint.ms_files instead!')
        self.ms_files = list_of_files

    @property
    def n_files(self):
        return len(self.ms_files)
    
    @ms_files.setter
    def ms_files(self, list_of_files):
        if isinstance(list_of_files, str):
            list_of_files = [list_of_files]
        list_of_files = [str(P(i)) for i in list_of_files if is_ms_file(i)]
        for f in list_of_files:
            if not os.path.isfile(f): 
                logging.warning(f'File not found ({f})')
        self._files = list_of_files
        if self.verbose:
            print( 'Set files to:\n' + '\n'.join(self.ms_files) + '\n' )

    @property
    def peaklist_files(self):
        return self._peaklist_files

    @peaklist_files.setter
    def peaklist_files(self, list_of_files):
        if isinstance(list_of_files, str):
            list_of_files = [list_of_files]
        if not isinstance(list_of_files, list):
            raise ValueError('Input should be a list of files.')
        for f in list_of_files:
            assert os.path.isfile(f), f'File not found ({f})' 
        self._peaklist_files = list_of_files
        if self.verbose: print( 'Set peaklist files to:\n'.join(self.peaklist_files) + '\n')
        self.peaklist = read_peaklists(list_of_files)

    @property
    def n_peaklist_files(self):
        return len(self.peaklist_files)
    
    @property
    def peaklist(self):
        return self._peaklist

    @peaklist.setter
    def peaklist(self, peaklist):
        peaklist = standardize_peaklist(peaklist)
        errors = check_peaklist(peaklist)
        if len(errors) != 0:
            peaklist = peaklist.drop_duplicates()
            error_string = '\n'.join(errors)
            if self.verbose:
                logging.error(f'Errors in peaklist:\n{error_string}')
            self._messages = errors
        self._peaklist = peaklist
        if self.verbose:
            print('Set peaklists to:\n', self.peaklist.to_string(), '\n')

    @property
    def results(self):
        return self._results
    
    @results.setter
    def results(self, df):
        self._results = df

    @property
    def rt_projections(self):
        return DeprecationWarning('rt_projections is deprecated. Peak shapes are now '
                                  'directly stored in the results table (mint.results).')
        
    def crosstab(self, col_name='peak_area'):
        return pd.crosstab(self.results.ms_file, 
                           self.results.peak_label, 
                           self.results[col_name], 
                           aggfunc=sum).astype(np.float64)
    @property
    def progress_callback(self):
        return self._progress_callback
    
    @progress_callback.setter
    def progress_callback(self, func=None):
        self._progress_callback = func
    
    @property
    def progress(self):
        return self._progress
    
    @progress.setter
    def progress(self, value):
        assert value >= 0, value
        assert value <= 100, value
        self._progress = value
        if self.progress_callback is not None:
            self.progress_callback(value)

    def export(self, filename=None):
        fn = filename
        if fn is None:
            buffer = export_to_excel(self, fn=fn)
            return buffer
        elif fn.endswith('.xlsx'):
            export_to_excel(self, fn=fn)
        elif fn.endswith('.csv'):
            self.results.to_csv(fn, index=False)

    def load(self, fn):
        if self.verbose: print('Loading MINT state')
        if isinstance(fn, str):
            if fn.endswith('xlsx'):
                results = pd.read_excel(fn, sheet_name='Results').rename(columns=DEPRECATED_LABELS)
                self.results = results
                self.peaklist = pd.read_excel(fn, sheet_name='Peaklist')
                self.ms_files = get_ms_files_from_results(results)

            elif fn.endswith('.csv'):
                results = pd.read_csv(fn).rename(columns=DEPRECATED_LABELS)
                ms_files = get_ms_files_from_results(results)
                peaklist = results[
                        [col for col in PEAKLIST_COLUMNS if col in results.columns]
                    ].drop_duplicates()
                self.results = results
                self.ms_files = ms_files
                self.peaklist = peaklist
                return None
        else:
            results = pd.read_csv(fn).rename(columns=DEPRECATED_LABELS)
            if 'ms_file' in results.columns:
                ms_files = get_ms_files_from_results(results)
                self.results = results
                self.ms_files = ms_files
            peaklist = results[[col for col in PEAKLIST_COLUMNS if col in results.columns]].drop_duplicates()
            self.peaklist = peaklist

    
    def plot_clustering(self, data=None, title=None, figsize=(8,8), target_var='peak_max',
                        vmin=-3, vmax=3, xmaxticks=None, ymaxticks=None, transpose=False,
                        normalize_by_row=True, transform_func='log1p', 
                        transform_filenames_func='basename', **kwargs):
        '''
        Performs a cluster analysis and plots a heatmap. If no data is provided, 
        data is taken form self.crosstab(target_var).
        First the data is transformed with the transform function (default is log(x+1)).
        If normalize_by_row==True the (transformed) data is also row-normalized (z=(x-μ)/σ)).
        A hierachical cluster analysis is performed with the (transformed) data.
        The original data is then re-ordered accoring to the resulting clusters.
        The result is stored in self.clustered.
        '''
        if len(self.results) == 0:
            return None
        simplefilter("ignore", ClusterWarning)
        if data is None:
            data = self.crosstab(target_var).copy()

        tmp_data = data.copy()

        if transform_func == 'log1p':
            transform_func = np.log1p

        if transform_func is not None:
            tmp_data = tmp_data.apply(transform_func)

        if transform_filenames_func == 'basename':
            transform_filenames_func = os.path.basename

        if transform_filenames_func is not None:
            tmp_data.columns = [transform_filenames_func(i) for i in tmp_data.columns] 

        if normalize_by_row:
            tmp_data = ((tmp_data.T - tmp_data.T.mean()) / tmp_data.T.std())

        if transpose: tmp_data = tmp_data.T    

        clustered, fig, ndx_x, ndx_y = hierarchical_clustering( 
            tmp_data, vmin=vmin, vmax=vmax, figsize=figsize, 
            xmaxticks=xmaxticks, ymaxticks=ymaxticks, **kwargs )

        if not transpose:
            self.clustered = data.iloc[ndx_y, ndx_x]
        else:
            self.clustered = data.iloc[ndx_x, ndx_y]
        return fig
    

    def plot_peak_shapes(self,  **kwargs):
        if len(self.results) > 0:
            return plot_peak_shapes(self.results,  **kwargs)


    def plot_heatmap(self, col_name='peak_max', normed_by_cols=False, transposed=False, 
            clustered=False, add_dendrogram=False, name='', correlation=False):
        '''Creates an interactive heatmap 
        that can be used to explore the data interactively.
        `mint.crosstab()` is called and then subjected to
        the `mint.vis.plotly.plotly_tools.plot_heatmap()`.

        Arguments
        ---------
        col_name: str, default='peak_max'
            Name of the column in `mint.results` to be analysed.
        normed_by_cols: bool, default=True
            Whether or not to normalize the columns in the crosstab.           
        clustered: bool, default=False
            Whether or not to cluster the rows. 
        add_dendrogram: bool, default=False
            Whether or not to replace row labels with a dendrogram.
        transposed: bool, default=False
            If True transpose matrix before plotting.
        correlation: bool, default=False
            If True convert data to correlation matrix before plotting.

        '''
        if len(self.results) > 0:
            return plot_heatmap(self.crosstab(col_name), normed_by_cols=normed_by_cols, 
                transposed=transposed, clustered=clustered, add_dendrogram=add_dendrogram, 
                name=col_name, correlation=correlation)


    def pca(self, var_name='peak_max', n_vars=20, fillna=0, 
            scaler='standard'):

        df = self.crosstab(var_name).fillna(fillna)
        
        if fillna == 'median':
            fillna = df.median()
        elif fillna == 'mean':
            fillna = df.mean()

        df = df.fillna(fillna)

        if scaler is not None:
            if scaler == 'standard':
                scaler = StandardScaler()
            elif scaler == 'robust':
                scaler = RobustScaler()
            df = pd.DataFrame(scaler.fit_transform(df), 
                    index=df.index, columns=df.columns)

        min_dim = min(df.shape)
        n_vars = min(n_vars, min_dim)
        pca = PCA(n_vars)
        X_projected = pca.fit_transform(df)
        df_projected = pd.DataFrame(X_projected, 
            index=df.index.get_level_values(0)).add_prefix('PCA-')

        explained_variance = pca.explained_variance_ratio_*100
        cumm_expl_var = np.cumsum(explained_variance)

        self.decomposition_results = {
            'df_projected': df_projected,
            'cumm_expl_var': cumm_expl_var,
            'n_vars': n_vars,
            'type': 'PCA'
        }


    def pca_plot_cummulative_variance(self):
        n_vars = self.decomposition_results['n_vars']
        fig = plt.figure(figsize=(7,3))
        cumm_expl_var = self.decomposition_results['cumm_expl_var']
        plt.bar(np.arange(n_vars)+1, cumm_expl_var, 
            facecolor='grey', edgecolor='none')
        plt.xlabel('# PCA-components')
        plt.ylabel('Explained variance')
        plt.title('Cummulative explained variance')
        plt.grid()
        return fig


    def plot_pair_plot(self, n_vars=3, color_groups=None, group_name=None):
        df = self.decomposition_results['df_projected']
        cols = df.columns.to_list()[:n_vars]
        df = df[cols]
        if color_groups is not None:
            if group_name is None: group_name = 'Group'
            df[group_name] = color_groups
            df[group_name] = df[group_name].astype(str)
        fig = plt.figure()
        g = sns.pairplot(df, plot_kws={'s': 100}, hue=group_name)
        if color_groups is not None:
            leg = g._legend
            leg.set_bbox_to_anchor([1.05, 0.5])
        return fig



from ms_mint.Mint import Mint

def test__run_skips_broken_files():
    mint = Mint(verbose=True)
    mint.peaklist_files = 'tests/data/peaklist_v0.csv'
    mint.files = ['tests/data/test.mzXML', 
                  'tests/data/test-broken.mzXML']
    mint.run()
    print(mint.results)


    broken_files_absent = all(mint.results.ms_file != 'test-broken.mzXML')
    good_files_present = all(mint.results.ms_file == 'test.mzXML')
    assert broken_files_absent, mint.results.ms_file
    assert good_files_present, mint.results.ms_file




import logging
import dash
import dash_html_components as html
import dash_bootstrap_components as dbc

from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State, ALL

_label = None

_layout = html.Div([
    html.Div(id='message')
])

def layout():
    return _layout

def callbacks(app, fsc, cache):

    @app.callback(
        Output('message', 'children'),
        [Input({'type': 'output', 'index': ALL}, 'children')],
        prevent_initial_call=True
        )

    def message(message):
        #print(f'Message: {message}')
        ctx = dash.callback_context
        message = ctx.triggered[0]
        if (message is None) or (message['value'] is None):
            #print(message)
            raise PreventUpdate
        _type = message['value']['type']
        _chil = message['value']['props']['children']
        #print(f'Message ({_type}):', _chil)
        if 'color' in message['value']['props'].keys():
            color = message['value']['props']['color']
        else: color = None
        if _type == 'Alert':
            return dbc.Alert(_chil, color=color)
        else:
            logging.warning(f'Undefined message type: {_type}, {_chil}')
        return dbc.Alert('Undefined message', 'danger')


import seaborn as sns
import pandas as pd



def plot_peak_shapes(mint_results, ms_files=None, peak_labels=None, height=4, aspect=1, legend=False,
                     n_cols=None, col_wrap=4, hue='ms_file', top=None, title=None, **kwargs):
    
    R = mint_results.copy()
    
    if peak_labels is not None:
        if isinstance(peak_labels, str):
            peak_labels = [peak_labels]
        R = R[R.peak_label.isin(peak_labels)]
    else:
        peak_labels = R.peak_label.drop_duplicates().values

    if ms_files is not None:
        R = R[R.ms_file.isin(ms_files)]
    
    dfs = []
    for peak_label in peak_labels:
        for ndx, row in R[(R.peak_label == peak_label) & (R.peak_n_datapoints>1)].iterrows():
            peak_rt = [float(i) for i in row.peak_shape_rt.split(',')]
            peak_int = [float(i) for i in row.peak_shape_int.split(',')]
            ms_file = row.ms_file
            df = pd.DataFrame({'Retention Time [min]': peak_rt, 'MS-Intensity': peak_int, 
                               'ms_file': ms_file, 'peak_label': peak_label})
            dfs.append(df)
            
    df = pd.concat(dfs)
    
    if n_cols is not None:
        col_wrap = n_cols

    fig = sns.relplot(
        data=df,
        x="Retention Time [min]", y="MS-Intensity",
        hue=hue, col="peak_label",
        kind="line", col_wrap=col_wrap,
        height=height, aspect=aspect, 
        facet_kws=dict(sharex=False, sharey=False),
        legend=legend,
        **kwargs
        )

    fig.set_titles(row_template = '{row_name}', col_template = '{col_name}')
    
    for ax in fig.axes.flatten():
        ax.ticklabel_format(style='sci', scilimits=(0,0), axis='both')
        
    if title is not None:
        fig.fig.suptitle(title, y=1.01)

    return fig


from tqdm import tqdm

import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt
import seaborn as sns  

import dash_html_components as html
import dash_bootstrap_components as dbc
import dash_core_components as dcc
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State

from . import tools as T

graph_options = [{'label': 'Histograms', 'value': 'hist'},
                 {'label': 'Boxplots', 'value': 'boxplot'},
                 {'label': 'Probability Density', 'value': 'density'}]

_label = 'Distributions'

_layout = html.Div([
    html.H3('Quality Control'),
    html.Button('Update', id='dist-update'),
    dcc.Dropdown(id='dist-graphs', options=graph_options, value=['hist', 'boxplot', 'density'], multi=True, placeholder='Kinds of graphs'),
    dcc.Checklist(id='dist-select', options=[{'label': 'Dense', 'value': 'Dense'}], value=['Dense']),
    html.Div(id='dist-figures', style={'float': 'center'})
])


layout_no_data = html.Div([
    dcc.Markdown('''### No results generated yet. 
    MINT has not been run yet. The Quality Control tabs uses the processed data. 
    To generate it please add MS-files as well as a valid peaklist. 
    Then execute MINT data processing routine witby clicking the `RUN MINT` button. 
    Once results have been produced you can access the QC tools.'''),
])

def layout():
    return _layout
    
def callbacks(app, fsc, cache):

    @app.callback(
    Output('dist-figures', 'children'),
    Input('dist-update', 'n_clicks'),
    State('tab', 'value'),
    State('ana-groupby', 'value'),
    State('dist-graphs', 'value'),
    State('dist-select', 'value'),
    State('ana-file-types', 'value'),
    State('ana-peak-labels-include', 'value'),
    State('ana-peak-labels-exclude', 'value'),
    State('wdir', 'children')
    )
    def qc_figures(n_clicks, tab, groupby, kinds, options, file_types, 
            include_labels, exclude_labels, wdir):

        if n_clicks is None:
            raise PreventUpdate

        df = T.get_complete_results( wdir, include_labels=include_labels, 
                exclude_labels=exclude_labels, file_types=file_types )

        if len(df) == 0: return 'No results yet. First run MINT.'

        if 'boxplot' in kinds:
            if (groupby is None or len(groupby) ==0):
                return dbc.Alert('For boxplots a "Group-by" column has to be set', color='info')
            if len(df[groupby].drop_duplicates()) <= 1:
                return dbc.Alert('For boxplots at least two groups have to be defined in selected "Group-by" column in metadata sheet.', color='info')

        sns.set_context('paper')
        
        sort_by_col = 'Batch'
        quant_col = 'log(peak_max+1)'

        if sort_by_col is not None:
            df = df.sort_values(['peak_label', sort_by_col])

        if options is None:
            options = []

        figures = []
        n_total = len(df.peak_label.drop_duplicates())
        for i, (peak_label, grp) in tqdm( enumerate(df.groupby('peak_label')), total=n_total ):

            if not 'Dense' in options: figures.append(dcc.Markdown(f'#### `{peak_label}`', style={'float': 'center'}))
            fsc.set('progress', int(100*(i+1)/n_total))

            # Sorting to ensure similar legends
            if sort_by_col is not None:
                grp = grp.sort_values(sort_by_col).reset_index(drop=True)

            if len(grp) < 1:
                continue

            if 'hist' in kinds: 
                sns.displot(data=grp, x=quant_col, height=3, hue=groupby, aspect=1)
                plt.title(peak_label)
                fig_label = f'by-{groupby}__{quant_col}__{peak_label}'
                T.savefig(kind='hist', wdir=wdir, label=fig_label)
                src = T.fig_to_src(dpi=150)
                figures.append( html.Img(src=src, style={'width': '300px'}) )

            if 'density' in kinds:
                sns.displot(data=grp, x=quant_col, hue=groupby, kind='kde', 
                            common_norm=False, height=3,  aspect=1)
                plt.title(peak_label)
                fig_label = f'by-{groupby}__{quant_col}__{peak_label}'
                T.savefig(kind='density', wdir=wdir, label=fig_label)                
                src = T.fig_to_src(dpi=150)
                figures.append( html.Img(src=src, style={'width': '300px'}) )

            if 'boxplot' in kinds:
                n_groups = len( grp[groupby].drop_duplicates() )
                aspect = max(1, n_groups/10)
                sns.catplot(data=grp, y=quant_col, x=groupby, height=3, kind='box', aspect=aspect, color='w')
                if quant_col in ['peak_max', 'peak_area']:
                    plt.gca().ticklabel_format(axis='y', style='sci', scilimits=(0,0))
                plt.title(peak_label)
                plt.xticks(rotation=90)
                fig_label = f'by-{groupby}__{quant_col}__{peak_label}'
                T.savefig(kind='boxplot', wdir=wdir, label=fig_label)                   
                src = T.fig_to_src(dpi=150)
                figures.append( html.Img(src=src, style={'width': '300px'}) )

            if not 'Dense' in options: figures.append(dcc.Markdown('---'))

            #if i == 3: break

        return figures



# ms_mint/processing.py

import os
import pandas as pd
import numpy as np

from .io import ms_file_to_df

from .standards import RESULTS_COLUMNS,\
    MINT_RESULTS_COLUMNS


def extract_chromatogram_from_ms1(df, mz_mean, mz_width, unit='minutes'):
    dmz = mz_mean*1e-6*mz_width
    chrom = df[(df['m/z array']-mz_mean).abs()<=dmz].copy()
    chrom['retentionTime'] = chrom['retentionTime'].round(3)
    chrom = chrom.groupby   ('retentionTime').max()
    return chrom['intensity array'] 


def process_ms1_files_in_parallel(args):
    '''
    Pickleable function for (parallel) peak integration.
    Expects a dictionary with keys:
        Mandatory:
        - 'filename': 'path to file to be processed',
        - 'peaklist': 'dataframe containing the peaklist' 
        - 'mode': 'express' or 'standard'
            * 'express' omits calculcation of rt projections
        Optional:
        - 'queue': instance of multiprocessing.Manager().Queue()

    Returns tuple with two elements:
        1) results, dataframe with integration results
        2) rt_projection, dictionary of dictionaries with peak shapes
    '''

    filename = args['filename']
    peaklist = args['peaklist']
    
    if 'queue' in args.keys():
        q = args['queue']
        q.put('filename')
    try:
        return process_ms1_file(filename=filename, peaklist=peaklist)
    except:
        return pd.DataFrame()


def process_ms1_file(filename, peaklist):
    '''
    Peak integration using a filename as input.
    -----
    Args:
        - filename: str or PosixPath, path to mzxml or mzml filename
        - peaklist: pandas.DataFrame(), DataFrame in peaklist format
    Returns:
        pandas.DataFrame(), DataFrame with processd peak intensities
    '''
    df = ms_file_to_df(filename)
    results = process_ms1(df, peaklist)
    results['total_intensity'] = df['intensity array'].sum()
    results['ms_file'] = os.path.basename(filename)
    results['ms_path'] = os.path.dirname(filename)
    results['ms_file_size'] = os.path.getsize(filename) / 1024 / 1024
    results['peak_score'] = score_peaks(results)
    return results[MINT_RESULTS_COLUMNS]


def process_ms1(df, peaklist):
    results = process_ms1_from_df(df, peaklist)
    results = pd.DataFrame(results, columns=['peak_label']+RESULTS_COLUMNS)
    results = pd.merge(peaklist, results, on=['peak_label'])
    results = results.reset_index(drop=True)
    return results


def process_ms1_from_df(df, peaklist):
    peak_cols = ['mz_mean', 'mz_width', 'rt_min', 'rt_max', 
                 'intensity_threshold', 'peak_label']
    array_peaks = peaklist[peak_cols].values
    array_data = df[['retentionTime', 'm/z array', 'intensity array']].values
    result = process_ms1_from_numpy(array_data, array_peaks)
    return result


def process_ms1_from_numpy(array, peaks):
    results = []
    for (mz_mean, mz_width, rt_min, rt_max, 
         intensity_threshold, peak_label) in peaks:
        props = _process_ms1_from_numpy(array, mz_mean=mz_mean, 
                    mz_width=mz_width, rt_min=rt_min, rt_max=rt_max, 
                    intensity_threshold=intensity_threshold, 
                    peak_label=peak_label)
        if props is None:
            continue
        results.append(
            [ props[col] for col in ['peak_label']+RESULTS_COLUMNS ]
        )
    return results


def _process_ms1_from_numpy(array, mz_mean, mz_width, rt_min, rt_max, 
                              intensity_threshold, peak_label=None):
    _slice = slice_ms1_array(array=array, mz_mean=mz_mean, mz_width=mz_width,
                             rt_min=rt_min, rt_max=rt_max, 
                             intensity_threshold=intensity_threshold)
    props = extract_ms1_properties(_slice, mz_mean)
    if props is None:
        return
    if peak_label is not None:
        props['peak_label'] = peak_label
    return props


def extract_ms1_properties(array, mz_mean):

    float_list_to_comma_sep_str = \
        lambda x: ','.join( [ str(np.round(i, 4)) for i in x ] )
    int_list_to_comma_sep_str = \
        lambda x: ','.join( [ str(int(i)) for i in x ] )
    
    projection = pd.DataFrame( array[:,[0,2]], columns=['rt', 'int'])
    projection['rt'] = projection['rt'].round(2)
    projection['int'] = projection['int'].astype(int)
    projection = projection.groupby('rt').max().reset_index().values

    times = array[:,0]
    masses = array[:,1]
    intensities = array[:,2]
    peak_n_datapoints = len(array)
    if peak_n_datapoints == 0:
        return None

    peak_area = intensities.sum()
    peak_mean = intensities.mean()
    peak_max = intensities.max()
    peak_min = intensities.min()
    peak_median = np.median(intensities)

    peak_rt_of_max = times[masses.argmax()]

    peak_delta_int = np.abs(intensities[0] - intensities[-1])

    peak_mass_diff_25pc, peak_mass_diff_50pc, peak_mass_diff_75pc = \
        np.quantile( masses, [.25,.5,.75] )
    
    peak_mass_diff_25pc -= mz_mean
    peak_mass_diff_50pc -= mz_mean
    peak_mass_diff_75pc -= mz_mean

    peak_mass_diff_25pc /= 1e-6*mz_mean
    peak_mass_diff_50pc /= 1e-6*mz_mean
    peak_mass_diff_75pc /= 1e-6*mz_mean

    peak_shape_rt = float_list_to_comma_sep_str( projection[:,0] )
    peak_shape_int = int_list_to_comma_sep_str( projection[:,1] )
    
    return dict(peak_area=peak_area, peak_max=peak_max, 
                peak_min=peak_min, peak_mean=peak_mean, 
                peak_rt_of_max=peak_rt_of_max, peak_median=peak_median,
                peak_delta_int=peak_delta_int, 
                peak_n_datapoints=peak_n_datapoints,
                peak_mass_diff_25pc=peak_mass_diff_25pc, 
                peak_mass_diff_50pc=peak_mass_diff_50pc, 
                peak_mass_diff_75pc=peak_mass_diff_75pc,
                peak_shape_rt=peak_shape_rt, 
                peak_shape_int=peak_shape_int,
                peak_score=None)


def slice_ms1_array(array: np.array, rt_min, rt_max, mz_mean, mz_width, 
                    intensity_threshold):
    delta_mass = mz_width*mz_mean*1e-6
    array = array[(array[:, 0] >= rt_min)]
    array = array[(array[:, 0] <= rt_max)]
    array = array[(np.abs(array[:, 1]-mz_mean) <= delta_mass)]
    array = array[(array[:, 2] >= intensity_threshold)]
    return array


#def gaus(x,a,x0,sigma):
#    return a*np.exp(-(x-x0)**2/(2*sigma**2))

def score_peaks(mint_results):
    R = mint_results.copy()
    scores = (((1-R.peak_delta_int.apply(abs)/R.peak_max)) 
              * ( np.tanh(R.peak_n_datapoints/20) )
              * ( 1/(1+abs(R.peak_rt_of_max - R[['rt_min', 'rt_max']].mean(axis=1) )) ))
    return scores




import os

import tempfile

import pandas as pd

from pathlib import Path as P

import matplotlib
matplotlib.use('Agg')

import dash_bootstrap_components as dbc
import dash_html_components as html
import dash_core_components as dcc

import dash
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
from dash_extensions import Download
from dash_extensions.enrich import FileSystemCache

from flask_caching import Cache

from . import tools as T

from . import workspaces
from . import ms_files
from . import metadata
from . import peaklist
from . import peak_optimization
from . import run_mint
from . import add_metab
from . import analysis
from . import messages

import dash_uploader as du
from tempfile import gettempdir

def make_dirs():
    tmpdir = tempfile.gettempdir()
    tmpdir = os.path.join(tmpdir, 'MINT')
    tmpdir = os.getenv('MINT_DATA_DIR', default=tmpdir)
    cachedir = os.path.join(tmpdir, '.cache')
    os.makedirs(tmpdir, exist_ok=True)
    os.makedirs(cachedir, exist_ok=True)
    return tmpdir, cachedir

TMPDIR, CACHEDIR = make_dirs()

fsc = FileSystemCache(CACHEDIR)

config = {
    "DEBUG": True,          # some Flask specific configs
    "CACHE_TYPE": "simple", # Flask-Caching related configs
    "CACHE_DEFAULT_TIMEOUT": 300
}

pd.options.display.max_colwidth = 1000

'''
components = {
    'workspaces':   {'label': 'Workspace',          'callbacks_func': workspaces.callbacks,         'layout_func': workspaces.layout},
    'msfiles':      {'label': 'MS-files',           'callbacks_func': ms_files.callbacks,           'layout_func': ms_files.layout},
    'metadata':     {'label': 'Metadata',           'callbacks_func': metadata.callbacks,           'layout_func': metadata.layout},
    'peaklist':     {'label': 'Peaklist',           'callbacks_func': peaklist.callbacks,           'layout_func': peaklist.layout},
    'add_metab':    {'label': 'Add Metabolites',    'callbacks_func': add_metab.callbacks,          'layout_func': add_metab.layout},
    'pko':          {'label': 'Peak Optimization',  'callbacks_func': peak_optimization.callbacks,  'layout_func': peak_optimization.layout},
    'run':          {'label': 'Run MINT',           'callbacks_func': run_mint.callbacks,           'layout_func': run_mint.layout},
    'analysis':     {'label': 'Analysis',           'callbacks_func': analysis.callbacks,           'layout_func': analysis.layout},
}
'''

_modules = [
  workspaces,
  ms_files,
  metadata,
  peaklist,
  add_metab,
  peak_optimization,
  run_mint,
  analysis
]

modules = {module._label: module for module in _modules}

# Collect outputs:
_outputs = html.Div(id='outputs', children=[module._outputs for module in 
                                     _modules if module._outputs is not None], 
             style={'visibility': 'hidden'})

app = dash.Dash(__name__, 
    external_stylesheets=[
        #dbc.themes.BOOTSTRAP,
        dbc.themes.MINTY,
        "https://codepen.io/chriddyp/pen/bWLwgP.css"
        ],
    requests_pathname_prefix=os.getenv('MINT_SERVE_PATH', default='/'),
    routes_pathname_prefix=os.getenv('MINT_SERVE_PATH', default='/')
    )

app.css.config.serve_locally = True
app.scripts.config.serve_locally = True


UPLOAD_FOLDER_ROOT = gettempdir()
du.configure_upload(app, UPLOAD_FOLDER_ROOT)


cache = Cache(app.server, config={
    'CACHE_TYPE': 'filesystem',
    'CACHE_DIR': '/tmp/MINT-cache'
})

app.title = 'MINT'

app.config['suppress_callback_exceptions'] = True

app.layout = html.Div([
    #html.Img(src=app.get_asset_url('logo.png'), style={'height': '30px'}),

    dcc.Interval(id="progress-interval", n_intervals=0, interval=500, disabled=False),

    html.A(href='https://soerendip.github.io/ms-mint/gui/', 
         children=[html.Button('Documentation', id='B_help', style={'float': 'right', 'color': 'info'})],
         target="_blank"),

    html.A(href=f'https://github.com/soerendip/ms-mint/issues/new?body={T.get_issue_text()}', 
         children=[html.Button('Issues', id='B_issues', style={'float': 'right', 'color': 'info'})],
         target="_blank"),

    dbc.Progress(id="progress-bar", value=100, style={'marginBottom': '20px', 'width': '100%'}),

    messages.layout(),

    Download(id='res-download-data'),


    html.Div(id='tmpdir', children=TMPDIR, style={'visibility': 'hidden'}),

    html.P('Current Workspace: ', style={'display': 'inline-block', 'marginRight': '5px', 'marginTop': '5px'}),

    html.Div(id='active-workspace', style={'display': 'inline-block'}),

    html.Div(id='wdir', children='', style={'display': 'inline-block', 'visibility': 'visible', 'float': 'right'}),

    html.Div(id='pko-creating-chromatograms'),

    dcc.Tabs(id='tab', value=_modules[0]._label,
        children=[
            dcc.Tab(id=modules[key]._label,
                    value=key, 
                    label=modules[key]._label,
                    )
            for key in modules.keys()]
    ),

    html.Div(id='pko-image-store', style={'visibility': 'hidden', 'height': '0px'}),

    html.Div(id='tab-content'),

    _outputs
    
], style={'margin':'2%'})

messages.callbacks(app=app, fsc=fsc, cache=cache)

for module in _modules:
    func = module.callbacks
    if func is not None:
        func(app=app, fsc=fsc, cache=cache)


@app.callback(
    Output('tab-content', 'children'),
    Input('tab', 'value'),
    State('wdir', 'children')
)
def render_content(tab, wdir):
    func = modules[tab].layout
    print('Working dir:', wdir)
    if tab != 'Workspaces' and wdir == '':
        return dbc.Alert('Please, create and activate a workspace.', color='warning')
    elif tab in ['Metadata', 'Peak Optimization', 'Processing'] and len(T.get_ms_fns( wdir )) == 0:
        return dbc.Alert('Please import MS files.', color='warning')
    elif tab in ['Processing'] and ( len(T.get_peaklist( wdir )) == 0 ):
        return dbc.Alert('Please, define peaklist.', color='warning')
    elif tab in ['Analysis'] and not P(T.get_results_fn( wdir )).is_file():
        return dbc.Alert('Please, create results (Processing).', color='warning')
    if func is not None:
        return func()
    else:
        raise PreventUpdate


if __name__ == '__main__':
    app.run_server(debug=True, threaded=True, 
        dev_tools_hot_reload_interval=5000,
        dev_tools_hot_reload_max_retry=30)



import os

import pandas as pd

import dash
import dash_bootstrap_components as dbc
import dash_html_components as html
import dash_core_components as dcc
from dash_table import DataTable
from dash.dependencies import Input, Output, State

from . import tools as T

from ms_mint.standards import PEAKLIST_COLUMNS
from ms_mint.peaklists import read_peaklists


columns = [{"name": i, "id": i, 
            "selectable": True}  for i in PEAKLIST_COLUMNS]

pkl_table = DataTable(
                id='pkl-table',
                columns=columns,
                data=None,
                sort_action="native",
                sort_mode="single",
                row_selectable=False,
                row_deletable=True,
                editable=True,
                column_selectable=False,
                selected_rows=[],
                page_action="native",
                page_current= 0,
                page_size= 30,
                filter_action='native',                
                style_table={'overflowX': 'scroll'},
                style_as_list_view=True,
                style_cell={'padding-left': '20px', 
                            'padding-right': '20px'},
                style_header={'backgroundColor': 'white',
                              'fontWeight': 'bold'},
                export_format='csv',
                export_headers='display',
                merge_duplicate_headers=True
            ) 

_label = 'Peaklist'

_layout = html.Div([
    html.H3('Peaklist'),

    dcc.Upload(
            id='pkl-upload',
            children=html.Div([
                'Drag and Drop or ',
                html.A('Select Files')
            ]),
            style={
                'width': '100%',
                'height': '60px',
                'lineHeight': '60px',
                'borderWidth': '1px',
                'borderStyle': 'dashed',
                'borderRadius': '5px',
                'textAlign': 'center',
                'margin': '10px'
            },
            # Allow multiple files to be uploaded
            multiple=True
        ),
    dcc.Dropdown('pkl-ms-mode', options=[
        {'value': 'positive', 'label': 'Add proton mass to formula (positive mode)'},
        {'value': 'negative', 'label': 'Subtract proton mass from formula (negative mode)'}], value=None),

    html.Button('Save', id='pkl-save'),
    html.Button('Clear', id='pkl-clear', style={'float': 'right'}),
    html.Div(id='table-container', children=pkl_table, style={'margin': '20px'})
])


_outputs = html.Div(id='pkl-outputs', children=[
    html.Div(id={'index': 'pkl-upload-output', 'type': 'output'}),
    html.Div(id={'index': 'pkl-save-output', 'type': 'output'}),
    html.Div(id={'index': 'pkl-clear-output', 'type': 'output'}),
])


def layout():
    return _layout


def callbacks(app, fsc=None, cache=None):
    @app.callback(
    Output('pkl-table', 'data'),
    Input('pkl-upload', 'contents'),
    Input('pkl-ms-mode', 'value'),
    Input('pkl-clear', 'n_clicks'),    
    State('pkl-upload', 'filename'),
    State('pkl-upload', 'last_modified'),
    State('wdir', 'children')
    )
    def pkl_upload(list_of_contents, ms_mode, clear, list_of_names, list_of_dates, wdir):
        prop_id = dash.callback_context.triggered[0]['prop_id']
        if prop_id.startswith('pkl-clear'):
            return pd.DataFrame(columns=PEAKLIST_COLUMNS).to_dict('records')
        target_dir = os.path.join(wdir, 'peaklist')
        fn = os.path.join( target_dir, 'peaklist.csv')
        if list_of_contents is not None:
            dfs = [T.parse_pkl_files(c, n, d, target_dir, ms_mode=ms_mode) for c, n, d in
                zip(list_of_contents, list_of_names, list_of_dates) ]
            data = dfs[0].to_dict('records')    
            return data
        elif os.path.isfile(fn):
            return read_peaklists(fn).to_dict('records')   

    @app.callback(
        Output({'index': 'pkl-save-output', 'type': 'output'}, 'children'),
        Input('pkl-save', 'n_clicks'),
        Input('pkl-table', 'data'),
        State('wdir', 'children')
    )
    def plk_save(n_clicks, data, wdir):
        df = pd.DataFrame(data)
        if len(df) == 0:
            df = pd.DataFrame(columns=PEAKLIST_COLUMNS)
        T.write_peaklist( df, wdir)
        return dbc.Alert('Peaklist saved.', color='info')







import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
from scipy.signal import find_peaks, peak_widths
from sklearn.neighbors import KernelDensity

from ..tools import gaussian
from ..Resampler import Resampler

class RetentionTimeOptimizer():
    def __init__(self, rt_min=None, rt_max=None, rt=None, rt_margin=None,
                 how='largest', dt=0.01, show_figure=False, precission=3
                 ):
        self.t_peaks = None
        self.x_peaks = None
        self.peak_width_bottom = None
        self.rt_max = rt_max
        self.rt_min = rt_min
        self.rt_margin = rt_margin
        self.rt = rt
        self.how = how
        self.dt = dt
        self.resampler = Resampler()
        self.show_figure = show_figure
        self.precission = precission
        self.update_values()


    def update_values(self):
        if (self.rt_min is not None) and (self.rt_max is not None):
            if self.rt is None:
                self.rt = np.mean([self.rt_min, self.rt_max])
            if self.rt_margin is None:
                self.rt_margin = (self.rt_max - self.rt_min) / 2
        
    def find_peaks_and_widths(self, chrom, prominence=1000):
        x = chrom.values
        ndx_peaks, _ = find_peaks(x, prominence=(prominence, None))
        t_peaks = chrom.iloc[ndx_peaks].index
        x_peaks = chrom.iloc[ndx_peaks].values
        results_bottom  = peak_widths(x, ndx_peaks, rel_height=0.90)
        self.t_peaks = t_peaks
        self.x_peaks = x_peaks
        self.peak_width_bottom = results_bottom
        
    def weighted_chrom(self, chrom):
        mean = self.rt
        sigma = self.rt_margin
        weights = gaussian(chrom.index, mean, sigma)
        return chrom*weights

    def ndx_largest_peak(self):
        x_peaks = self.x_peaks
        if len(x_peaks) == 0:
            return None
        return np.argmax(x_peaks)
        
    def resample(self, chrom):
        return self.resampler.resample(chrom)
    
    def get_chrom_properties(self, chrom):
        self.t_min = chrom.index.min()
        self.t_max = chrom.index.max()
        self.t_span = self.t_max - self.t_min
        self.len_chrom = len(chrom)
    
    def x_to_t(self, x):
        return ((self.t_span * np.array(x) / self.len_chrom) + self.t_min).flatten()
    
    def find_largest_peak_in_chromatogram(self, chrom):
        chrom = chrom.copy()
        chrom = self.weighted_chrom(chrom)
        if self.show_figure: chrom.plot(color='grey', alpha=0.5)
        chrom = self.resample(chrom)
        self.get_chrom_properties(chrom)        
        if self.show_figure: chrom.plot(ax=plt.gca(), color='darkblue', alpha=0.5)
        self.find_peaks_and_widths(chrom)
        ndx = self.ndx_largest_peak()
        if ndx is None: return None
        prop = self.get_peak_properties(ndx)
        if self.show_figure: 
            plt.xlim(prop[2]-0.1, prop[3]+0.1)
            plt.hlines(-0.1, prop[2], prop[3], lw=2)
        return prop
    
    def find_largest_peak(self, chromatograms):
        props = []
        for chrom in chromatograms:
            prop = self.find_largest_peak_in_chromatogram(chrom)
            if prop is not None: props.append( prop )
        if len(props) == 0: return (None, None)
        df = pd.DataFrame( props, columns=['rt', 'max_intensity', 'rt_min', 'rt_max'] )

        rt_of_largest_peak_max = df.sort_values('max_intensity', ascending=False).iloc[0]['rt'] 
        # Filter out far away peaks
        df = df[(df.rt - rt_of_largest_peak_max).abs()<0.3]
        rt_min = estimate_expectation_value(df.rt_min)
        rt_max = estimate_expectation_value(df.rt_max)
        if self.show_figure:
            plt.vlines([rt_min, rt_max], 0, np.max(df.max_intensity), label='Selected RT range', color='k', ls='--')
        return np.round(rt_min, self.precission), np.round(rt_max, self.precission)
    
    def get_peak_start_timee(self, ndx):
        return self.x_to_t(self.peak_width_bottom[2][ndx])[0]
 
    def get_peak_end_timee(self, ndx):
        return self.x_to_t(self.peak_width_bottom[3][ndx])[0]
    
    def get_peak_properties(self, ndx):
        return (self.t_peaks[ndx], 
                self.x_peaks[ndx], 
                self.get_peak_start_timee(ndx),
                self.get_peak_end_timee(ndx))


def estimate_expectation_value(values, kernel='gaussian'):
    values = np.array(values)
    assert kernel in ['gaussian', 'tophat', 'epanechnikov']
    x = np.linspace(np.min(values)-1, np.max(values)+1, 1000)[:, np.newaxis]
    kde = KernelDensity(kernel=kernel, bandwidth=0.5).fit(values[:, np.newaxis])
    log_dens = kde.score_samples(x)
    ndx = np.argmax(log_dens)
    return x[ndx][0]



import pandas as pd

import dash
import dash_html_components as html
import dash_core_components as dcc
import dash_bootstrap_components as dbc
from dash_extensions.javascript import Namespace
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from dash_tabulator import DashTabulator

from . import tools as T


ns = Namespace("myNamespace", "tabulator")

tabulator_options = {
           "groupBy": "Label", 
           "selectable": True,
           "headerFilterLiveFilterDelay":3000,
           #"dataSorted" : ns("dataSorted"),
           "layout": "fitDataFill",
           "height": "900px",
           }

downloadButtonType = {"css": "btn btn-primary", "text":"Export", "type":"csv", "filename":"Metadata"}
clearFilterButtonType = {"css": "btn btn-outline-dark", "text":"Clear Filters"}

meta_table = html.Div(id='meta-table-container', 
    style={'minHeight':  100, 'margin': '0%'},
    children=[
        DashTabulator(id='meta-table',
            columns=T.gen_tabulator_columns(add_ms_file_col=True, add_color_col=True, add_peakopt_col=True), 
            options=tabulator_options,
            downloadButtonType=downloadButtonType,
            clearFilterButtonType=clearFilterButtonType
        )
])

options = [
    {'label': 'Batch', 'value': 'Batch'},
    {'label': 'Label', 'value': 'Label'},
    {'label': 'Color', 'value': 'Color'},
    {'label': 'Type', 'value': 'Type'},
    {'label': 'Concentration', 'value': 'Concentration'},
    ]

_label = 'Metadata'

_layout = html.Div([
    html.H3('Metadata'),
    dcc.Upload(
            id='meta-upload',
            children=html.Div([
                'Drag and Drop or ',
                html.A('Select Files')
            ]),
            style={
                'width': '100%',
                'height': '60px',
                'lineHeight': '60px',
                'borderWidth': '1px',
                'borderStyle': 'dashed',
                'borderRadius': '5px',
                'textAlign': 'center',
                'margin': '10px'
            },
            # Allow multiple files to be uploaded
            multiple=True,
        ),
    dcc.Markdown('---'),
    dcc.Markdown('##### Actions'),
    dbc.Row([
        dcc.Dropdown(
            id='meta-action', options=[
                {'label': 'Set', 'value': 'Set'},
                {'label': 'Create column', 'value': 'Create column'},
                {'label': 'Delete column', 'value': 'Delete column'}
            ],
            value='Set', style={'width': '150px'}),
        dcc.Dropdown(
            id='meta-column', options=options, 
            value=None, style={'width': '150px'}),
        dcc.Input(id='meta-input'),
        dcc.Dropdown(
            id='meta-input-bool', options=[{'value': 'True', 'label': 'True'},
                                           {'value': 'False', 'label': 'False'}],
            value=None),

        html.Button('Apply', id='meta-apply'),
    ], style={'marginLeft': '5px'}),

    html.Div(id='meta-table-saved-on-edit-output'),
    dcc.Markdown('---'),    
    dcc.Loading( meta_table ),
])


_outputs = html.Div(id='meta-outputs', 
    children=[
        html.Div(id={'index': 'meta-apply-output', 'type': 'output'})
    ]
)


def layout():
    return _layout


def callbacks(app, fsc, cache):
        
    @app.callback(
    Output('meta-table', 'data'),
    Output('meta-table', 'columns'),
    Output('meta-column', 'options'),
    Input('meta-upload', 'contents'),
    Input('meta-upload', 'filename'),
    Input({'index': 'meta-apply-output', 'type': 'output'}, 'children'),
    State('wdir', 'children')
    )
    def meta_upload(contents, filename, message, wdir):
        metadata = T.get_metadata( wdir )
        if (contents is not None) and (len(contents) > 0):
            contents = T.parse_table_content(contents[0], filename[0])
            metadata = T.merge_metadata(metadata, contents)
        columns = metadata.columns.to_list()
        columns = [{'label':col, 'value':col} for col in columns if col != 'index']
        if 'index' not in metadata.columns: metadata = metadata.reset_index()
        return metadata.to_dict('records'), T.gen_tabulator_columns(metadata.columns,
            add_ms_file_col=True, add_color_col=True, add_peakopt_col=True), columns


    @app.callback(
    Output({'index': 'meta-apply-output', 'type': 'output'}, 'children'),
    Input('meta-apply', 'n_clicks'),
    State('meta-table', 'multiRowsClicked'),
    State('meta-table', 'data'),
    State('meta-table', 'dataFiltered'),
    State('meta-action', 'value'),
    State('meta-column', 'value'),
    State('meta-input', 'value'),
    State('meta-input-bool', 'value'),
    State('wdir', 'children'),
    )
    def meta_save(n_clicks, selected_rows, data, 
                  data_filtered, action, column, value, 
                  value_bool, wdir):
        
        if action == 'Set' and column == 'PeakOpt':
            value = value_bool

        if data is None or len(data) == 0:
            raise PreventUpdate

        df = pd.DataFrame(data)

        if 'index' in df.columns:
            df = df.set_index('index')
        else:
            df = df.reset_index()
        prop_id = dash.callback_context.triggered[0]['prop_id']
        if prop_id == 'meta-apply.n_clicks':
            if action == 'Set':
                filtered_rows = [r for r in data_filtered['rows'] if r is not None]
                filtered_ndx = [r['index'] for r in filtered_rows]
                if selected_rows == []:
                    # If nothing is selected apply to all visible rows
                    ndxs = filtered_ndx
                else:
                    # If something is selected only apply to selected rows
                    ndxs = [r['index'] for r in selected_rows if r['index'] in filtered_ndx]
                if len(ndxs) == 0 or column is None:
                    return dbc.Alert('No rows selected.', color='danger')
                df.loc[ndxs, column] = value
            elif action == 'Create column': df[value] = ''
            elif action == 'Delete column': del df[column]

        T.write_metadata(df, wdir )

        if prop_id == 'meta-table.cellEdited':
            raise PreventUpdate
        return dbc.Alert('Metadata saved.', color='info')

    
    @app.callback(
        Output('meta-table-saved-on-edit-output', 'children'),
        Input('meta-table', 'cellEdited'),
        State('meta-table', 'data'),
        State('wdir', 'children'),
    )
    def save_table_on_edit(cell_edited, data, wdir):
        if data is None or cell_edited is None:
            raise PreventUpdate
        df = pd.DataFrame(data)
        T.write_metadata( df , wdir)
    
    
    @app.callback(
        Output('meta-column', 'disabled'),
        Input('meta-action', 'value'),
    )
    def set_something(action):
        if action in ['Set', 'Delete column']:
            return False
        return True
    
    
    @app.callback(
        Output('meta-input', 'style'),
        Output('meta-input-bool', 'style'),
        Input('meta-action', 'value'),
        Input('meta-column', 'value')
    )
    def set_something(action, column): 
        visible = {'visibility': 'visible', 'width': '150px', 'margin': 0}
        hidden = {'visibility': 'hidden', 'width': '0px', 'margin':0}
        if (action == 'Set') & (column == 'PeakOpt'):
            return hidden, visible
        return visible, hidden
        


import re
import shutil

import pandas as pd

import dash
import dash_html_components as html
import dash_core_components as dcc
import dash_bootstrap_components as dbc

from dash.exceptions import PreventUpdate

from dash_table import DataTable
from dash.dependencies import Input, Output, State

from . import tools as T


ws_table = html.Div(id='ws-table-container', 
    style={'minHeight':  100, 'margin': '5%'},
    children=[
        DataTable(id='ws-table',
            columns=[ {"name": i, "id": i, "selectable": True}  
                                for i in ['Workspace']],
                    data=[],
                    row_selectable='single',
                    row_deletable=False,
                    style_cell={'textAlign': 'left'},
                    sort_action='native'
                    )
])


_layout = html.Div([
    html.H3('Workspaces'),
    html.Button('Create Workspace', id='ws-create'),
    html.Button('Delete Workspace', id='ws-delete', style={'float': 'right'}),

    ws_table,

    dbc.Modal(
        [dbc.ModalHeader("Create Workspace"),
         dbc.ModalBody(
             [dcc.Input(id='ws-create-input', placeholder='New workspace name'),
              html.P(id='ws-create-output')
             ]
         ),
         dbc.ModalFooter(
            html.Div([
                dbc.Button("Create", id="ws-create-confirm", style={'marginRight': '10px'}),
                dbc.Button("Cancel", id="ws-create-cancel")
            ])
        )], id="ws-create-popup"
    ),

    dbc.Modal(
            [
                dbc.ModalHeader("Delete Workspace"),
                dbc.ModalBody("This will delete all files and results in the selected workspace."),
                dbc.ModalFooter(
                    html.Div([
                        dbc.Button("Delete", id="ws-delete-confirm", style={'marginRight': '10px'}),
                        dbc.Button("Cancel", id="ws-delete-cancel")
                    ])
                ),
            ],
            id="ws-delete-popup",
        ),
])


_outputs = html.Div(id='ws-outputs', 
    children=[
        html.Div(id={'index': 'ws-created-output',  'type': 'output'}),
        html.Div(id={'index': 'ws-activate-output', 'type': 'output'}),
        html.Div(id={'index': 'ws-delete-output',   'type': 'output'})
    ]
)

_label = 'Workspaces'

def layout():
    return _layout

def callbacks(app, fsc, cache):

    @app.callback(
    Output('ws-table', 'data'),
    Input({'index': 'ws-created-output', 'type': 'output'}, 'children'),
    Input('tab', 'value'),
    Input({'index': 'ws-delete-output', 'type': 'output'}, 'children'),
    State('tmpdir', 'children')
    )
    def ws_table(value, tab, delete, tmpdir):
        T.maybe_migrate_workspaces(tmpdir)
        ws_names = T.get_workspaces(tmpdir)
        ws_names.sort()
        ws_names = [{'Workspace': ws_name} for ws_name in ws_names 
                        if not ws_name.startswith('.')]
        return ws_names




    @app.callback(
    Output({'index': 'ws-activate-output', 'type': 'output'}, 'children'),
    Output('wdir', 'children'),
    Output('active-workspace', 'children'),
    Input('ws-table', 'derived_virtual_selected_rows'),
    Input({'index': 'ws-delete-output', 'type': 'output'}, 'children'),
    Input({'index': 'ws-created-output', 'type': 'output'}, 'children'),
    State('ws-table', 'data'),
    State('tmpdir', 'children')
    )
    def ws_activate(ndx, deleted, created, data, tmpdir):
        prop_id = dash.callback_context.triggered[0]['prop_id']        
        if tmpdir is None: raise PreventUpdate
        ws_names = T.get_workspaces( tmpdir )
        if ws_names is None or len(ws_names)==0 : 
            message = 'No workspace defined.'
            return dbc.Alert(message, color='danger'), '', ''
        if prop_id == 'ws-delete-output.children' or ndx is None:
            ndx = [0]

        data = pd.DataFrame(data) 
        if len(ndx) == 1:
            ws_name = data.loc[ndx[0], 'Workspace']
        else:
            ws_name = T.get_active_workspace( tmpdir )

        if ws_name is None: raise PreventUpdate
        wdir = T.workspace_path(tmpdir, ws_name)
        if ws_name is not None: T.save_activated_workspace(tmpdir, ws_name)
        else: raise PreventUpdate
        message = f'Workspace {ws_name} activated.'
        if ws_name is None: ws_name = ''
        return dbc.Alert(message, color='success'), wdir, ws_name


    @app.callback(
    Output('progress-bar', 'value'),
    Input('progress-interval', 'n_intervals'),
    )
    def set_progress(n):
        return fsc.get('progress')


    @app.callback(
    Output("ws-delete-popup", "is_open"),
    Input("ws-delete", "n_clicks"), 
    Input("ws-delete-cancel", "n_clicks"),
    Input('ws-delete-confirm', 'n_clicks'),
    State("ws-delete-popup", "is_open")
    )
    def ws_delete(n1, n2, n3, is_open):
        if n1 is None:
            raise PreventUpdate
        if n1 or n2 or n3:
            return not is_open
        return is_open



    @app.callback(
        Output({'index': 'ws-delete-output', 'type': 'output'}, 'children'),
        Input('ws-delete-confirm', 'n_clicks'),
        State('ws-table', 'derived_virtual_selected_rows'),
        State('ws-table', 'data'),
        State('tmpdir', 'children')
    )
    def ws_delete_confirmed(n_clicks, ndxs, data, tmpdir):
        if n_clicks is None or len(ndxs) == 0:
            raise PreventUpdate
        for ndx in ndxs:
            ws_name = data[ndx]['Workspace']
            dirname = T.workspace_path( tmpdir, ws_name )
            shutil.rmtree(dirname)
        message = f'Worskpace {ws_name} deleted.'
        return dbc.Alert(message, color='info')



    @app.callback(
    Output("ws-create-popup", "is_open"),
    Input("ws-create", "n_clicks"), 
    Input("ws-create-cancel", "n_clicks"),
    Input('ws-create-confirm', 'n_clicks'),
    State("ws-create-popup", "is_open")
    )
    def ws_create(n1, n2, n3, is_open):
        if n1 is None:
            raise PreventUpdate
        if n1 or n2 or n3:
            return not is_open
        return is_open    


    @app.callback(
        Output('ws-create-output', 'children'),
        Output('ws-create-confirm', 'disabled'),
        Input('ws-create-input', 'value'),
        State('tmpdir', 'children')
    )
    def ws_create_message(ws_name, tmpdir):
        if (ws_name is None):
            raise PreventUpdate
        elif ws_name == '':
            return dbc.Alert('Name cannot be empty', color='warning'), True
        elif not re.match('^[\w_-]+$', ws_name):
            return dbc.Alert('Name can only contain: a-z, A-Z, 0-9, -,  _ and no blanks.', color='warning'), True
        elif T.workspace_exists(tmpdir, ws_name):
            return dbc.Alert('Workspace already exists', color='warning'), True
        else:
            return dbc.Alert(f'Can create workspace "{ws_name}"', color='success'), False


    @app.callback(
        Output({'index': 'ws-created-output', 'type': 'output'}, 'children'),
        Input('ws-create-confirm', 'n_clicks'),
        State('ws-create-input', 'value'),
        State('tmpdir', 'children')
    )
    def ws_create_confirmed(n_clicks, ws_name, tmpdir):
        if n_clicks is None: raise PreventUpdate
        T.create_workspace(tmpdir, ws_name)
        return dbc.Alert(f'Workspace {ws_name} created.', color='success')


    @app.callback(
        Output('ws-table', 'selected_rows'),
        Input('tab', 'value'),
        Input('ws-table', 'data'),
        State('tmpdir', 'children')
    )
    def set_selected_row(tab, data, tmpdir):
        if tab != _label: raise PreventUpdate
        ws_names = T.get_workspaces(tmpdir)
        if len(ws_names) == 0: raise PreventUpdate
        data = pd.DataFrame(data)
        active_ws = T.get_active_workspace(tmpdir)
        if active_ws is None:
            ndx = data.index[0]
            active_ws = data.Workspace[0]
        else:
            ndx = ws_names.index(active_ws)            
        if ndx is None:
            ndx = 0
        return [ndx]


import numpy as np
import pandas as pd
import ipywidgets as W
import plotly.express as px

from tqdm import tqdm
from IPython.display import display


from .io import ms_file_to_df

class ManualRetentionTimeOptimizer():

    def __init__(self, mint):
        
        self.df = pd.concat( [ms_file_to_df(fn).assign(ms_file=fn) for fn in tqdm(mint.ms_files)] )
        
        self.out = W.Output()

        self.mint = mint
        
        self.w_rt_min = W.FloatText(
            value=0,
            description='RT min:',
            disabled=False
        )

        self.w_rt_max = W.FloatText(
            value=13,
            description='RT max:',
            disabled=False,
        )

        self.set_rt_button = W.Button(description='Set new RT')
        self.delete_button = W.Button(description='Remove from peaklist')

        self.menu = W.Dropdown(options=mint.peaklist.peak_label, value=None)

        def update(*args):
            peak_label = self.menu.value    
            self.plot(peak_label)

        def update_rt(button):
            rt_min, rt_max = self.w_rt_min.value, self.w_rt_max.value, 
            peak_label = self.menu.value
            self.mint.peaklist.loc[
                self.mint.peaklist.peak_label==peak_label, 'rt_min'] = rt_min
            self.mint.peaklist.loc[
                self.mint.peaklist.peak_label==peak_label, 'rt_max'] = rt_max
            self.plot(peak_label)    

        def remove_peak(button):
            peak_label = self.menu.value
            mint.peaklist = mint.peaklist[mint.peaklist.peak_label!=peak_label]
            new_options = mint.peaklist.peak_label
            self.menu.options = new_options

        self.menu.observe(update, names='value')
        self.set_rt_button.on_click(update_rt)
        self.delete_button.on_click(remove_peak)

        self.layout = W.VBox( [self.menu,
                               self.w_rt_min, 
                               self.w_rt_max, 
                               self.set_rt_button,         
                               self.out, 
                               self.delete_button],
                                
                               )

    def plot(self, peak_label):
        peak_data = self.mint.peaklist[self.mint.peaklist.peak_label==peak_label].T.iloc[:,0]
        mz_mean, mz_width, rt_min, rt_max = peak_data[['mz_mean','mz_width', 'rt_min', 'rt_max']]
        dmz = mz_mean*1e-6*mz_width
        selection = self.df[np.abs(self.df['m/z array'] - mz_mean)<=dmz]
        fig = px.line( data_frame=selection, x='retentionTime', 
                       y='intensity array', color='ms_file', title=peak_label)
        fig.update_layout(showlegend=False)
        fig.update_layout(
            hovermode= 'closest',        
            xaxis=dict(
                range=[rt_min, rt_max]
            ))

        self.out.clear_output()
        with self.out:
            display( fig )

        self.w_rt_min.value, self.w_rt_max.value = rt_min, rt_max

    def show(self):
        return self.layout


import numpy as np
from molmass import Formula

from .standards import M_PROTON


def get_mz_mean_from_formulas(formulas, ms_mode=None, verbose=False):
    if verbose: print(formulas)
    masses = []
    for formula in formulas:
        if verbose: print(f'Mass from formula: "{formula}"')
        try:
            mass = Formula(formula).isotope.mass  
        except:
            masses.append(None)
            continue
        if ms_mode == 'positive':
            mass += M_PROTON
        elif ms_mode == 'negative':
            mass -= M_PROTON
        mass = np.round(mass, 4)
        masses.append(mass)
    if verbose: print(masses)
    return masses


def gaussian(x, mu, sig):
    return np.exp(-np.power(x - mu, 2.) / (2 * np.power(sig, 2.)))